import math
import os
import warnings

import matplotlib
import numpy as np
from echo import delay_callback
from astropy.visualization import ManualInterval, ContrastBiasStretch
from glue.core.subset_group import GroupedSubset
from glue.config import stretches as glue_stretches
from glue.viewers.scatter.state import ScatterViewerState
from glue.viewers.profile.state import ProfileViewerState, ProfileLayerState
from glue.viewers.image.state import ImageSubsetLayerState, ImageViewerState
from glue.viewers.scatter.state import ScatterLayerState as BqplotScatterLayerState
from glue.viewers.image.composite_array import COLOR_CONVERTER
from glue_jupyter.bqplot.image.state import BqplotImageLayerState
from glue_jupyter.common.toolbar_vuetify import read_icon
from scipy.interpolate import PchipInterpolator
from traitlets import Any, Dict, Float, Bool, Int, List, Unicode, observe

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, ViewerSelectMixin, LayerSelect,
                                        PlotOptionsSyncState, Plot,
                                        skip_if_no_updates_since_last_active, with_spinner)
from jdaviz.core.events import ChangeRefDataMessage
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.custom_traitlets import IntHandleEmpty
# by importing from utils, glue_colormaps will include the custom Random colormap
from jdaviz.utils import is_not_wcs_only, cmap_samples, glue_colormaps

__all__ = ['PlotOptions']

RANDOM_SUBSET_SIZE = 10_000


class SplineStretch:
    """
    A class to represent spline stretches.

    Attributes
    ----------
    k : int
        Degree of the smoothing spline. Default is 3.
    bc_type : str or None
        Boundary condition type. Default is None.
    t : array-like or None
        Array of knot positions. Default is None.
    x : array-like
        The x-coordinates of the data points.
    y : array-like
        The y-coordinates of the data points.
    spline : object
        Interpolating spline.

    Raises
    ------
    ValueError
        If `x` and `y` have different lengths.
    """

    def __init__(self):
        # Default x, y values(0-1) range chosen for a typical initial spline shape.
        # Can be modified if required.
        self._x = np.array([0, 0.1, 0.2, 0.7, 1])
        self._y = np.array([0, 0.05, 0.3, 0.9, 1])
        self.update_knots(self._x, self._y)

    @property
    def knots(self):
        return (self._x, self._y)

    @knots.setter
    def knots(self, value):
        x, y = value
        if len(x) != len(y):
            # Silently return
            return
        self.update_knots(x, y)

    def __call__(self, values, out=None, clip=False):
        # For our uses, we can ignore `out` and `clip`, but those would need
        # to be implemented before contributing this class upstream.
        return self.spline(values)

    def update_knots(self, x, y):
        self._x = x
        self._y = y
        self.spline = PchipInterpolator(self._x, self._y)


# Add the spline stretch to the glue stretch registry if not registered
if "spline" not in glue_stretches:
    glue_stretches.add("spline", SplineStretch, display="Spline")


def _round_step(step):
    # round the step for a float input
    if step <= 0:
        return 1e-6, 6
    decimals = -int(np.log10(abs(step))) + 1 if step != 0 else 6
    if decimals < 0:
        decimals = 0
    return float(np.round(step, decimals)), decimals


@tray_registry('g-plot-options', label="Plot Options")
class PlotOptions(PluginTemplateMixin, ViewerSelectMixin):
    """
    The Plot Options Plugin gives access to per-viewer and per-layer options and enables
    setting across multiple viewers/layers simultaneously.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``viewer`` (:class:`~jdaviz.core.template_mixin.ViewerSelect`):
    * ``viewer_multiselect``
    * ``layer`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
    * ``layer_multiselect``
    * :meth:`select_all`
    * ``subset_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      whether a subset should be visible.
    * ``subset_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``subset_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``axes_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_width`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_as_steps`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``uncertainty_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``stretch_function`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_preset`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_vmin`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_vmax`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_hist_zoom_limits`` : whether to show the histogram for the current zoom
      limits instead of all data within the layer; not exposed for Specviz.
    * ``stretch_hist_nbins`` : number of bins to use in creating the histogram; not exposed
      for Specviz.
    * ``stretch_curve_visible`` : bool
      whether the stretch histogram's colormap "curve" is visible; not exposed for Specviz.
    * ``image_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      whether the image bitmap is visible; not exposed for Specviz.
    * ``image_color_mode`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``image_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz.  This only applies when ``image_color_mode`` is "Color".
    * ``image_colormap`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``image_color_mode`` is "Colormap".
    * ``image_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Valid values are between 0 and 1, inclusive. Default is 1.
    * ``image_contrast`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Valid values are between 0 and 4, inclusive. Default is 1.
    * ``image_bias`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Valid values are between 0 and 1, inclusive. Default is 0.5.
    * ``contour_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      whether the contour is visible; not exposed for Specviz
    * ``contour_mode`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``contour_min`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Linear".
    * ``contour_max`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Linear".
    * ``contour_nlevels`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Linear".
    * ``contour_custom_levels`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Custom".
    """
    template_file = __file__, "plot_options.vue"
    uses_active_status = Bool(True).tag(sync=True)

    # read-only display units
    display_units = Dict().tag(sync=True)

    viewer_limits = Dict().tag(sync=True)

    layer_multiselect = Bool(False).tag(sync=True)
    layer_items = List().tag(sync=True)
    layer_selected = Any().tag(sync=True)  # Any needed for multiselect

    # profile/line viewer/layer options:
    line_visible_value = Bool().tag(sync=True)
    line_visible_sync = Dict().tag(sync=True)

    line_color_value = Any().tag(sync=True)
    line_color_sync = Dict().tag(sync=True)

    line_width_value = Int().tag(sync=True)
    line_width_sync = Dict().tag(sync=True)

    line_opacity_value = Float().tag(sync=True)
    line_opacity_sync = Dict().tag(sync=True)

    line_as_steps_value = Bool().tag(sync=True)
    line_as_steps_sync = Dict().tag(sync=True)

    uncertainty_visible_value = Int().tag(sync=True)
    uncertainty_visible_sync = Dict().tag(sync=True)

    x_min_value = Float().tag(sync=True)
    x_min_sync = Dict().tag(sync=True)

    x_max_value = Float().tag(sync=True)
    x_max_sync = Dict().tag(sync=True)

    y_min_value = Float().tag(sync=True)
    y_min_sync = Dict().tag(sync=True)

    y_max_value = Float().tag(sync=True)
    y_max_sync = Dict().tag(sync=True)

    x_bound_step = Float(0.1).tag(sync=True)  # dynamic based on maximum value
    y_bound_step = Float(0.1).tag(sync=True)  # dynamic based on maximum value

    zoom_center_x_value = Float().tag(sync=True)
    zoom_center_x_sync = Dict().tag(sync=True)

    zoom_center_y_value = Float().tag(sync=True)
    zoom_center_y_sync = Dict().tag(sync=True)

    zoom_radius_value = Float().tag(sync=True)
    zoom_radius_sync = Dict().tag(sync=True)

    zoom_step = Float(1).tag(sync=True)

    # scatter/marker options
    marker_visible_value = Bool().tag(sync=True)
    marker_visible_sync = Dict().tag(sync=True)

    marker_fill_value = Bool().tag(sync=True)
    marker_fill_sync = Dict().tag(sync=True)

    marker_opacity_value = Float().tag(sync=True)
    marker_opacity_sync = Dict().tag(sync=True)

    marker_size_mode_value = Unicode().tag(sync=True)
    marker_size_mode_sync = Dict().tag(sync=True)

    marker_size_value = Float().tag(sync=True)
    marker_size_sync = Dict().tag(sync=True)

    marker_size_scale_value = Float().tag(sync=True)
    marker_size_scale_sync = Dict().tag(sync=True)

    marker_size_col_value = Unicode().tag(sync=True)
    marker_size_col_sync = Dict().tag(sync=True)

    marker_size_vmin_value = Float().tag(sync=True)
    marker_size_vmin_sync = Dict().tag(sync=True)

    marker_size_vmax_value = Float().tag(sync=True)
    marker_size_vmax_sync = Dict().tag(sync=True)

    marker_color_mode_value = Unicode().tag(sync=True)
    marker_color_mode_sync = Dict().tag(sync=True)

    marker_color_value = Any().tag(sync=True)
    marker_color_sync = Dict().tag(sync=True)

    marker_color_col_value = Unicode().tag(sync=True)
    marker_color_col_sync = Dict().tag(sync=True)

    marker_colormap_value = Unicode().tag(sync=True)
    marker_colormap_sync = Dict().tag(sync=True)

    marker_colormap_vmin_value = Float().tag(sync=True)
    marker_colormap_vmin_sync = Dict().tag(sync=True)

    marker_colormap_vmax_value = Float().tag(sync=True)
    marker_colormap_vmax_sync = Dict().tag(sync=True)

    # image viewer/layer options
    stretch_function_value = Unicode().tag(sync=True)
    stretch_function_sync = Dict().tag(sync=True)

    stretch_preset_value = Any().tag(sync=True)  # glue will pass either a float or string
    stretch_preset_sync = Dict().tag(sync=True)

    stretch_vstep = Float(0.1).tag(sync=True)  # dynamic based on full range from image

    stretch_vmin_value = Float().tag(sync=True)
    stretch_vmin_sync = Dict().tag(sync=True)

    stretch_vmax_value = Float().tag(sync=True)
    stretch_vmax_sync = Dict().tag(sync=True)

    stretch_params_value = Dict().tag(sync=True)
    stretch_params_sync = Dict().tag(sync=True)

    stretch_hist_sync = Dict().tag(sync=True)
    stretch_hist_zoom_limits = Bool().tag(sync=True)
    stretch_hist_nbins = IntHandleEmpty(25).tag(sync=True)
    stretch_histogram_widget = Unicode().tag(sync=True)

    stretch_curve_visible = Bool(True).tag(sync=True)

    subset_visible_value = Bool().tag(sync=True)
    subset_visible_sync = Dict().tag(sync=True)

    subset_color_value = Unicode().tag(sync=True)
    subset_color_sync = Dict().tag(sync=True)

    subset_opacity_value = Float().tag(sync=True)
    subset_opacity_sync = Dict().tag(sync=True)

    image_visible_value = Bool().tag(sync=True)
    image_visible_sync = Dict().tag(sync=True)

    image_color_mode_value = Unicode().tag(sync=True)
    image_color_mode_sync = Dict().tag(sync=True)

    image_color_value = Any().tag(sync=True)
    image_color_sync = Dict().tag(sync=True)

    image_colormap_value = Unicode().tag(sync=True)
    image_colormap_sync = Dict().tag(sync=True)

    image_opacity_value = Float().tag(sync=True)
    image_opacity_sync = Dict().tag(sync=True)

    image_contrast_value = Float().tag(sync=True)
    image_contrast_sync = Dict().tag(sync=True)

    image_bias_value = Float().tag(sync=True)
    image_bias_sync = Dict().tag(sync=True)

    contour_spinner = Bool().tag(sync=True)

    contour_visible_value = Bool().tag(sync=True)
    contour_visible_sync = Dict().tag(sync=True)

    contour_mode_value = Unicode().tag(sync=True)
    contour_mode_sync = Dict().tag(sync=True)

    contour_min_value = Float().tag(sync=True)
    contour_min_sync = Dict().tag(sync=True)

    contour_max_value = Float().tag(sync=True)
    contour_max_sync = Dict().tag(sync=True)

    contour_nlevels_value = Int().tag(sync=True)
    contour_nlevels_sync = Dict().tag(sync=True)

    contour_custom_levels_value = List().tag(sync=True)
    contour_custom_levels_txt = Unicode().tag(sync=True)   # controlled by vue
    contour_custom_levels_sync = Dict().tag(sync=True)

    axes_visible_value = Bool().tag(sync=True)
    axes_visible_sync = Dict().tag(sync=True)

    icon_radialtocheck = Unicode(read_icon(os.path.join(ICON_DIR, 'radialtocheck.svg'), 'svg+xml')).tag(sync=True)  # noqa
    icon_checktoradial = Unicode(read_icon(os.path.join(ICON_DIR, 'checktoradial.svg'), 'svg+xml')).tag(sync=True)  # noqa

    cmap_samples = Dict(cmap_samples).tag(sync=True)
    swatches_palette = List().tag(sync=True)
    apply_RGB_presets_spinner = Bool(False).tag(sync=True)
    stretch_hist_spinner = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Set viewer and layer display options.'

        self.layer = LayerSelect(self, 'layer_items', 'layer_selected',
                                 'viewer_selected', 'layer_multiselect')

        self.layer.filters += [is_not_wcs_only]

        self.swatches_palette = [
            ['#FF0000', '#AA0000', '#550000'],
            ['#FFD300', '#AAAA00', '#555500'],
            ['#4CFF00', '#00AA00', '#005500'],
            ['#00FF8E', '#00AAAA', '#005555'],
            ['#0089FF', '#5200FF', '#000055']
        ]

        def is_profile(state):
            return isinstance(state, (ProfileViewerState, ProfileLayerState))

        def not_profile(state):
            return not is_profile(state)

        def is_scatter(state):
            return isinstance(state, (ScatterViewerState, BqplotScatterLayerState))

        def supports_line(state):
            return is_profile(state) or is_scatter(state)

        def is_image(state):
            return isinstance(state, BqplotImageLayerState)

        def not_image(state):
            return not is_image(state)

        def not_image_viewer(state):
            return not isinstance(state, ImageViewerState)

        def not_image_or_spatial_subset(state):
            return not is_image(state) and not is_spatial_subset(state)

        def is_spatial_subset(state):
            return isinstance(state, ImageSubsetLayerState) and is_not_wcs_only(state.layer)

        def is_not_subset(state):
            return not is_spatial_subset(state)

        def line_visible(state):
            # exclude for scatter layers where the marker is shown instead of the line
            return getattr(state, 'line_visible', True)

        def state_attr_for_line_visible(state):
            if is_scatter(state):
                return 'line_visible'
            return 'visible'

        # Profile/line viewer/layer options:
        self.line_visible = PlotOptionsSyncState(self, self.viewer, self.layer, state_attr_for_line_visible,  # noqa
                                                 'line_visible_value', 'line_visible_sync',
                                                 state_filter=supports_line)
        self.line_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                               'line_color_value', 'line_color_sync',
                                               state_filter=not_image_or_spatial_subset)
        self.line_width = PlotOptionsSyncState(self, self.viewer, self.layer, 'linewidth',
                                               'line_width_value', 'line_width_sync',
                                               state_filter=supports_line)
        self.line_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                 'line_opacity_value', 'line_opacity_sync',
                                                 state_filter=supports_line)
        self.line_as_steps = PlotOptionsSyncState(self, self.viewer, self.layer, 'as_steps',
                                                  'line_as_steps_value', 'line_as_steps_sync')
        self.uncertainty_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'show_uncertainty',  # noqa
                                                        'uncertainty_visible_value', 'uncertainty_visible_sync')  # noqa

        # Viewer bounds
        self.x_min = PlotOptionsSyncState(self, self.viewer, self.layer, 'x_min',
                                          'x_min_value', 'x_min_sync',
                                          state_filter=not_image_viewer)
        self.x_max = PlotOptionsSyncState(self, self.viewer, self.layer, 'x_max',
                                          'x_max_value', 'x_max_sync',
                                          state_filter=not_image_viewer)
        self.y_min = PlotOptionsSyncState(self, self.viewer, self.layer, 'y_min',
                                          'y_min_value', 'y_min_sync',
                                          state_filter=not_image_viewer)
        self.y_max = PlotOptionsSyncState(self, self.viewer, self.layer, 'y_max',
                                          'y_max_value', 'y_max_sync',
                                          state_filter=not_image_viewer)
        self.zoom_center_x = PlotOptionsSyncState(self, self.viewer, self.layer, 'zoom_center_x',
                                                  'zoom_center_x_value', 'zoom_center_x_sync')
        self.zoom_center_y = PlotOptionsSyncState(self, self.viewer, self.layer, 'zoom_center_y',
                                                  'zoom_center_y_value', 'zoom_center_y_sync')
        self.zoom_radius = PlotOptionsSyncState(self, self.viewer, self.layer, 'zoom_radius',
                                                'zoom_radius_value', 'zoom_radius_sync')

        # Scatter/marker options:
        # NOTE: marker_visible hides the entire layer (including the line)
        self.marker_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'visible',
                                                   'marker_visible_value', 'marker_visible_sync',
                                                   state_filter=is_scatter)
        self.marker_fill = PlotOptionsSyncState(self, self.viewer, self.layer, 'fill',
                                                'marker_fill_value', 'marker_fill_sync',
                                                state_filter=is_scatter)
        self.marker_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                   'marker_opacity_value', 'marker_opacity_sync',
                                                   state_filter=is_scatter)
        self.marker_size_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_mode',
                                                     'marker_size_mode_value', 'marker_size_mode_sync',  # noqa
                                                     state_filter=is_scatter)
        self.marker_size = PlotOptionsSyncState(self, self.viewer, self.layer, 'size',
                                                'marker_size_value', 'marker_size_sync',
                                                state_filter=is_scatter)
        self.marker_size_scale = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_scaling',
                                                      'marker_size_scale_value', 'marker_size_scale_sync',  # noqa
                                                      state_filter=is_scatter)
        self.marker_size_col = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_att',
                                                    'marker_size_col_value', 'marker_size_col_sync',
                                                    state_filter=is_scatter)
        self.marker_size_vmin = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_vmin',
                                                     'marker_size_vmin_value', 'marker_size_vmin_sync',  # noqa
                                                     state_filter=is_scatter)
        self.marker_size_vmax = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_vmax',
                                                     'marker_size_vmax_value', 'marker_size_vmax_sync',  # noqa
                                                     state_filter=is_scatter)

        # TODO: remove marker_ prefix if these also apply to the lines?
        self.marker_color_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_mode',
                                                      'marker_color_mode_value', 'marker_color_mode_sync',  # noqa
                                                      state_filter=is_scatter)
        self.marker_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                                 'marker_color_value', 'marker_color_sync',
                                                 state_filter=is_scatter)
        self.marker_color_col = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_att',
                                                     'marker_color_col_value', 'marker_color_col_sync',  # noqa
                                                     state_filter=is_scatter)
        self.marker_colormap = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap',
                                                    'marker_colormap_value', 'marker_colormap_sync',
                                                    state_filter=is_scatter)
        self.marker_colormap_vmin = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_vmin',
                                                         'marker_colormap_vmin_value', 'marker_colormap_vmin_sync',  # noqa
                                                         state_filter=is_scatter)
        self.marker_colormap_vmax = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_vmax',
                                                         'marker_colormap_vmax_value', 'marker_colormap_vmax_sync',  # noqa
                                                         state_filter=is_scatter)

        # Image viewer/layer options:
        self.stretch_function = PlotOptionsSyncState(self, self.viewer, self.layer, 'stretch',
                                                     'stretch_function_value', 'stretch_function_sync',  # noqa
                                                     state_filter=is_image)
        # use add_observe to ensure that the glue state syncs with the traitlet choice:
        self.stretch_function.add_observe('stretch_function_value', self._update_stretch_curve)

        self.stretch_preset = PlotOptionsSyncState(self, self.viewer, self.layer, 'percentile',
                                                   'stretch_preset_value', 'stretch_preset_sync',
                                                   state_filter=is_image)
        self.stretch_vmin = PlotOptionsSyncState(self, self.viewer, self.layer, 'v_min',
                                                 'stretch_vmin_value', 'stretch_vmin_sync',
                                                 state_filter=is_image)
        self.stretch_vmax = PlotOptionsSyncState(self, self.viewer, self.layer, 'v_max',
                                                 'stretch_vmax_value', 'stretch_vmax_sync',
                                                 state_filter=is_image)
        self.stretch_params = PlotOptionsSyncState(self, self.viewer, self.layer, 'stretch_parameters',  # noqa
                                                   'stretch_params_value', 'stretch_params_sync',
                                                   state_filter=is_image)

        self.stretch_histogram = Plot(self, name='stretch_hist', viewer_type='histogram',
                                      update_callback=self._update_stretch_histogram)
        # Add the stretch bounds tool to the default Plot viewer.
        self.stretch_histogram.tools_nested.append(["jdaviz:stretch_bounds"])
        self.stretch_histogram._initialize_toolbar(["jdaviz:stretch_bounds"])

        self.stretch_histogram._add_data('histogram', x=[0, 1])

        self.stretch_histogram.add_line('vmin', x=[0, 0], y=[0, 1], ynorm=True, color='#c75d2c')
        self.stretch_histogram.add_line('vmax', x=[0, 0], y=[0, 1], ynorm='vmin', color='#c75d2c')
        self.stretch_histogram.add_line(
            label='stretch_curve',
            x=[], y=[],
            ynorm='vmin',
            color="#007BA1",  # "inactive" blue
            opacities=[0.5],
        )
        self.stretch_histogram.add_scatter(
            label='stretch_knots',
            x=[], y=[],
            ynorm='vmin',
            color="#c75d2c",  # "active" orange (tool enabled by default)
        )
        self.stretch_histogram.add_scatter('colorbar', x=[], y=[], ynorm='vmin', marker='square', stroke_width=33)  # noqa: E501
        self.stretch_histogram.viewer.state.update_bins_on_reset_limits = False
        self.stretch_histogram.viewer.state.x_limits_percentile = 95
        with self.stretch_histogram.figure.hold_sync():
            self.stretch_histogram.figure.axes[0].label = 'pixel value'
            self.stretch_histogram.figure.axes[0].num_ticks = 3
            self.stretch_histogram.figure.axes[0].tick_format = '0.1e'
            self.stretch_histogram.figure.axes[1].label = 'density'
            self.stretch_histogram.figure.axes[1].num_ticks = 2
        self.stretch_histogram_widget = f'IPY_MODEL_{self.stretch_histogram.model_id}'

        self.subset_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'visible',
                                                   'subset_visible_value', 'subset_visible_sync',
                                                   state_filter=is_spatial_subset)
        self.subset_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                                 'subset_color_value', 'subset_color_sync',
                                                 state_filter=is_spatial_subset)
        self.subset_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                   'subset_opacity_value', 'subset_opacity_sync',
                                                   state_filter=is_spatial_subset)
        self.image_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'bitmap_visible',
                                                  'image_visible_value', 'image_visible_sync',
                                                  state_filter=is_image)
        self.image_color_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'color_mode',  # noqa
                                                     'image_color_mode_value', 'image_color_mode_sync')  # noqa
        self.image_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                                'image_color_value', 'image_color_sync',
                                                state_filter=is_image)
        self.image_colormap = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap',
                                                   'image_colormap_value', 'image_colormap_sync')
        self.image_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                  'image_opacity_value', 'image_opacity_sync',
                                                  state_filter=is_image)
        self.image_contrast = PlotOptionsSyncState(self, self.viewer, self.layer, 'contrast',
                                                   'image_contrast_value', 'image_contrast_sync')
        self.image_bias = PlotOptionsSyncState(self, self.viewer, self.layer, 'bias',
                                               'image_bias_value', 'image_bias_sync')

        self.contour_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'contour_visible',  # noqa
                                                    'contour_visible_value', 'contour_visible_sync',
                                                    spinner='contour_spinner')
        self.contour_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'level_mode',
                                                 'contour_mode_value', 'contour_mode_sync',
                                                 spinner='contour_spinner')
        self.contour_min = PlotOptionsSyncState(self, self.viewer, self.layer, 'c_min',
                                                'contour_min_value', 'contour_min_sync',
                                                spinner='contour_spinner')
        self.contour_max = PlotOptionsSyncState(self, self.viewer, self.layer, 'c_max',
                                                'contour_max_value', 'contour_max_sync',
                                                spinner='contour_spinner')
        self.contour_nlevels = PlotOptionsSyncState(self, self.viewer, self.layer, 'n_levels',
                                                    'contour_nlevels_value', 'contour_nlevels_sync',
                                                    spinner='contour_spinner')
        self.contour_custom_levels = PlotOptionsSyncState(self, self.viewer, self.layer, 'levels',
                                                          'contour_custom_levels_value', 'contour_custom_levels_sync',   # noqa
                                                          spinner='contour_spinner')

        # Axes options:
        # axes_visible hidden for imviz in plot_options.vue
        self.axes_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'show_axes',
                                                 'axes_visible_value', 'axes_visible_sync',
                                                 state_filter=not_profile)
        sv = self.spectrum_viewer
        if sv is not None:
            sv.state.add_callback('x_display_unit',
                                  self._on_global_display_unit_changed)
            sv.state.add_callback('y_display_unit',
                                  self._on_global_display_unit_changed)

        self.hub.subscribe(self, ChangeRefDataMessage,
                           handler=self._on_refdata_change)

    @property
    def user_api(self):
        expose = ['multiselect', 'viewer', 'viewer_multiselect', 'layer', 'layer_multiselect',
                  'select_all', 'subset_visible', 'reset_viewer_bounds']
        if self.config == "cubeviz":
            expose += ['uncertainty_visible']
        if self.config != "imviz":
            expose += ['x_min', 'x_max', 'y_min', 'y_max',
                       'axes_visible', 'line_visible', 'line_color', 'line_width', 'line_opacity',
                       'line_as_steps', 'uncertainty_visible']
        if self.config != "specviz":
            expose += ['zoom_center_x', 'zoom_center_y', 'zoom_radius',
                       'subset_color', 'subset_opacity',
                       'stretch_function', 'stretch_preset', 'stretch_vmin', 'stretch_vmax',
                       'stretch_hist_zoom_limits', 'stretch_hist_nbins',
                       'image_visible', 'image_color_mode',
                       'image_color', 'image_colormap', 'image_opacity',
                       'image_contrast', 'image_bias',
                       'contour_visible', 'contour_mode',
                       'contour_min', 'contour_max', 'contour_nlevels', 'contour_custom_levels',
                       'stretch_curve_visible', 'apply_RGB_presets']

        return PluginUserApi(self, expose)

    @property
    def multiselect(self):
        warnings.warn(
            "multiselect has been replaced by separate viewer_multiselect and "
            "layer_multiselect and will be removed in the future. "
            "This currently evaluates viewer_multiselect or layer_multiselect", DeprecationWarning)
        return self.viewer_multiselect or self.layer_multiselect

    @multiselect.setter
    def multiselect(self, value):
        warnings.warn(
            "multiselect has been replaced by separate viewer_multiselect and "
            "layer_multiselect and will be removed in the future. "
            "This currently sets viewer_multiselect and layer_multiselect", DeprecationWarning)
        self.viewer_multiselect = value
        self.layer_multiselect = value

    def select_all(self, viewers=True, layers=True):
        """
        Enable multiselect mode and select all viewers and/or layers.

        Parameters
        ----------
        viewers : bool
            Whether to set ``viewer_multiselect`` and select all viewers (default: True)

        layers: bool
            Whether to set ``layer_multiselect`` and select all layers (default: True)
        """
        if viewers:
            self.viewer_multiselect = True
            self.viewer.select_all()
        if layers:
            self.layer_multiselect = True
            self.layer.select_all()

    def _on_global_display_unit_changed(self, *args):
        sv = self.spectrum_viewer
        self.display_units['spectral'] = sv.state.x_display_unit
        self.display_units['flux'] = sv.state.y_display_unit
        self.send_state('display_units')

    def _on_refdata_change(self, *args):
        if self.app._align_by.lower() == 'wcs':
            self.display_units['image'] = 'deg'
        else:
            self.display_units['image'] = 'pix'
        self.send_state('display_units')
        self._update_viewer_zoom_steps()

    def vue_unmix_state(self, names):
        if isinstance(names, str):
            names = [names]
        for name in names:
            sync_state = getattr(self, name)
            sync_state.unmix_state()
        if 'stretch_params' in names:
            # there is no way to call send_state to force the update to the layers,
            # so we'll force an update by clearing first
            stretch_params = dict(self.stretch_params_value)
            self.stretch_params_value = {}
            self.stretch_params_value = stretch_params

    def vue_set_value(self, data):
        attr_name = data.get('name')
        value = data.get('value')
        setattr(self, attr_name, value)

    @with_spinner('apply_RGB_presets_spinner')
    def apply_RGB_presets(self):
        """
        Applies preset colors, opacities, and stretch settings to all visible layers
        (in all viewers) when in Color (Monochromatic) mode.
        """

        if (self.image_color_mode_value != "One color per layer" or
                self.image_color_mode_sync['mixed']):
            raise ValueError("RGB presets can only be applied if color mode is Color.")
        # Preselected colors we want to use for 5 or less layers
        preset_colors = [self.swatches_palette[4][1],
                         "#0000FF",
                         "#00FF00",
                         self.swatches_palette[1][0],
                         self.swatches_palette[0][0],
                         ]

        preset_inds = {2: [1, 4], 3: [1, 2, 4], 4: [1, 2, 3, 4]}

        # Switch back to this at the end
        initial_layer = self.layer_selected

        # Determine layers visible in selected viewer(s) - consider mixed to be visible
        visible_layers = [layer['label'] for layer in self.layer.items if not layer['is_subset'] and (layer['visible'] in (True, 'mixed'))]  # noqa

        # Set opacity to something that seems sensible
        n_visible = len(visible_layers)
        default_opacity = 1
        if n_visible > 2:
            default_opacity = 1 / math.log2(n_visible)

        # Sample along a colormap if we have too many layers
        if n_visible > len(preset_colors):
            cmap = matplotlib.colormaps['gist_rainbow'].resampled(n_visible)
            # Have to reverse the order of the cmap to make physical sense with
            # assumed wavelength order of layers.
            preset_colors = [matplotlib.colors.to_hex(cmap(i), keep_alpha=True) for
                             i in range(n_visible - 1, -1, -1)]
        elif n_visible >= 2 and n_visible < len(preset_colors):
            preset_colors = [preset_colors[i] for i in preset_inds[n_visible]]

        for i in range(n_visible):
            self.layer_selected = visible_layers[i]
            self.image_opacity.unmix_state(default_opacity)
            self.image_color.unmix_state(preset_colors[i])
            self.stretch_function.unmix_state("arcsinh")
            self.stretch_preset.unmix_state(99)

        self.layer_selected = initial_layer

    def vue_apply_RGB_presets(self, data):
        self.apply_RGB_presets()

    @observe('viewer_selected',
             'x_min_value', 'x_max_value',
             'y_min_value', 'y_max_value')
    def _update_viewer_bound_steps(self, msg={}):
        if not hasattr(self, 'viewer'):  # pragma: no cover
            # plugin hasn't been fully initialized yet
            return

        if not self.viewer.selected or not self.x_min_sync['in_subscribed_states']:
            # nothing selected yet
            return

        for ax in ('x', 'y'):
            ax_min = getattr(self, f'{ax}_min_value')
            ax_max = getattr(self, f'{ax}_max_value')
            bound_step, decimals = _round_step((ax_max - ax_min) / 100.)
            decimals = -int(np.log10(abs(bound_step))) + 1 if bound_step != 0 else 6
            setattr(self, f'{ax}_bound_step', bound_step)
            setattr(self, f'{ax}_min_value', np.round(ax_min, decimals=decimals))
            setattr(self, f'{ax}_max_value', np.round(ax_max, decimals=decimals))

    @observe('viewer_selected',
             'zoom_center_x_value', 'zoom_center_y_value',
             'zoom_radius_value')
    def _update_viewer_zoom_steps(self, msg={}):
        if not hasattr(self, 'viewer'):  # pragma: no cover
            # plugin hasn't been fully initialized yet
            return

        if not self.viewer.selected or not self.zoom_radius_sync['in_subscribed_states']:
            # nothing selected yet
            return

        # in the case of multiple viewers, calculate based on the first
        # alternatively, we could find the most extreme by looping over all selected viewers
        viewers = self.viewer.selected_obj if self.viewer_multiselect else [self.viewer.selected_obj]  # noqa
        for viewer in viewers:
            if hasattr(viewer.state, '_get_reset_limits'):
                break
        else:
            # no image viewer
            return
        x_min, x_max, y_min, y_max = viewer.state._get_reset_limits(return_as_world=True)
        self.zoom_step, _ = _round_step(max(x_max-x_min, y_max-y_min) / 100.)

    def vue_reset_viewer_bounds(self, _):
        self.reset_viewer_bounds()

    def reset_viewer_bounds(self):
        # This button is currently only exposed if only the spectrum viewer is selected
        viewers = [self.viewer.selected_obj] if not self.viewer_multiselect else self.viewer.selected_obj # noqa
        for viewer in viewers:
            viewer.toolbar.tools['jdaviz:homezoom'].activate()

    @observe('stretch_function_sync', 'stretch_params_sync',
             'stretch_vmin_sync', 'stretch_vmax_sync',
             'image_color_mode_sync', 'image_color_sync', 'image_colormap_sync')
    def _update_stretch_hist_sync(self, msg={}):
        # the histogram should show as mixed if ANY of the input parameters are mixed
        # these should match in the @observe above, all_syncs here, as well as the strings
        # passed to unmix_state in the <glue-state-sync-wrapper> in plot_options.vue
        all_syncs = [self.stretch_function_sync, self.stretch_params_sync,
                     self.stretch_vmin_sync, self.stretch_vmax_sync,
                     self.image_color_mode_sync, self.image_color_sync, self.image_colormap_sync]
        self.stretch_hist_sync = {'in_subscribed_states': bool(np.any([sync.get('in_subscribed_states', False) for sync in all_syncs])),  # noqa
                                  'mixed': bool(np.any([sync.get('mixed', False) for sync in all_syncs]))}  # noqa

    @observe('is_active', 'layer_selected', 'viewer_selected',
             'stretch_hist_zoom_limits')
    @skip_if_no_updates_since_last_active()
    def _request_update_stretch_histogram(self, msg={}):
        if not hasattr(self, 'viewer'):  # pragma: no cover
            # plugin hasn't been fully initialized yet
            return

        # NOTE: this method is separate from _update_stretch_histogram so that
        # _update_stretch_histogram can be called manually (or from the
        # update_callback on the Plot object itself) without going through
        # the skip_if_no_updates_since_last_active check (and can therefore
        # be executed even if the plugin is not active)
        self._update_stretch_histogram(msg)

    def _zoom_limits_update_stretch_histogram(self, msg={}):
        if not self.stretch_hist_zoom_limits:
            # there isn't anything to update, let's not waste resources
            return
        self._update_stretch_histogram()

    @with_spinner('stretch_hist_spinner')
    def _update_stretch_histogram(self, msg={}):
        if not self.stretch_function_sync.get('in_subscribed_states'):  # pragma: no cover
            # no (image) viewer with stretch function options
            return

        if not self.viewer.selected or not self.layer.selected:  # pragma: no cover
            # nothing to plot, will be hidden in UI
            return

        if self.layer_multiselect and len(self.layer.selected) > 1:
            # currently only support single-layer, if multiple layers are selected, the plot
            # will be hidden in the UI
            return

        if not self._viewer_is_image_viewer():
            # don't update histogram if selected viewer is not an image viewer:
            return

        viewer = self.viewer.selected_obj[0] if self.viewer_multiselect else self.viewer.selected_obj  # noqa

        # manage viewer zoom limit callbacks
        if ((isinstance(msg, dict) and msg.get('name') == 'viewer_selected')
                or not self.stretch_hist_zoom_limits):
            vs = viewer.state
            for attr in ('x_min', 'x_max', 'y_min', 'y_max'):
                vs.add_callback(attr, self._zoom_limits_update_stretch_histogram)
        if isinstance(msg, dict) and msg.get('name') == 'viewer_selected':
            viewer_label_old = msg.get('old')
            if isinstance(viewer_label_old, list):
                viewer_label_old = viewer_label_old[0]
            # If the previously selected viewer was deleted, we don't need to do this.
            if viewer_label_old in self.app._viewer_store:
                vs_old = self.app.get_viewer(viewer_label_old).state
                for attr in ('x_min', 'x_max', 'y_min', 'y_max'):
                    vs_old.remove_callback(attr, self._zoom_limits_update_stretch_histogram)

        if not len(self.layer.selected_obj):
            # skip further updates if no data are available:
            return
        if isinstance(self.layer.selected_obj[0], list):
            if not len(self.layer.selected_obj[0]):
                return
            # multiselect case (but we won't check multiselect since the selection can lag behind)
            layer = self.layer.selected_obj[0][0]
        else:
            layer = self.layer.selected_obj[0]
        data = layer.layer

        if isinstance(data, GroupedSubset):
            # don't update histogram for subsets:
            return

        comp = data.get_component(layer.state.attribute)

        # TODO: further optimization could be done by caching sub_data
        if self.stretch_hist_zoom_limits and (not self.layer_multiselect or len(self.layer_selected) == 1):  # noqa
            if hasattr(viewer, '_get_zoom_limits'):
                # Viewer limits. This takes account of Imviz linking.
                xy_limits = viewer._get_zoom_limits(data).astype(int)
                x_limits = xy_limits[:, 0]
                y_limits = xy_limits[:, 1]
                x_min = max(x_limits.min(), 0)
                x_max = x_limits.max()
                y_min = max(y_limits.min(), 0)
                y_max = y_limits.max()

                sub_data = comp.data[y_min:y_max, x_min:x_max]

            else:
                # spectrum-2d-viewer, for example.  We'll assume the viewer
                # limits correspond to the fixed data components from glue
                # and filter directly.
                x_data = data.get_component(data.components[1]).data
                y_data = data.get_component(data.components[0]).data

                inverted_x = getattr(viewer, 'inverted_x_axis', False)
                x_min = viewer.state.x_min if not inverted_x else viewer.state.x_max
                x_max = viewer.state.x_max if not inverted_x else viewer.state.x_min
                inds = np.where((x_data >= x_min) &
                                (x_data <= x_max) &
                                (y_data >= viewer.state.y_min) &
                                (y_data <= viewer.state.y_max))

                sub_data = comp.data[inds]

        else:
            # include all data, regardless of zoom limits
            sub_data = comp.data

        self.stretch_histogram.viewer.state.random_subset = RANDOM_SUBSET_SIZE
        self.stretch_histogram._update_data('histogram', x=sub_data)

        if len(sub_data) > 0:

            # Use glue to compute the statistics since this allows us to use
            # a random subset of the data to compute the histogram.
            # The 2.5 and 97.5 hardcoded here is equivalent to
            # PercentileInterval(95).get_limits(sub_data)
            glue_data = self.stretch_histogram.app.data_collection['histogram']
            hist_lims = (
                glue_data.compute_statistic('percentile', glue_data.id['x'],
                                            percentile=2.5, random_subset=RANDOM_SUBSET_SIZE),
                glue_data.compute_statistic('percentile', glue_data.id['x'],
                                            percentile=97.5, random_subset=RANDOM_SUBSET_SIZE)
            )

            # set the stepsize for vmin/vmax to be approximately 1% of the range of the
            # histogram (within the percentile interval), rounded to 1-2 significant digits
            # to avoid random step sizes.  This logic is somewhat arbitrary and can be safely
            # modified or eventually exposed to the user if that would be useful.
            stretch_vstep = (hist_lims[1] - hist_lims[0]) / 100.
            self.stretch_vstep = _round_step(stretch_vstep)[0]

            with delay_callback(self.stretch_histogram.viewer.state, 'hist_x_min', 'hist_x_max'):
                self.stretch_histogram.viewer.state.hist_x_min = hist_lims[0]
                self.stretch_histogram.viewer.state.hist_x_max = hist_lims[1]

        self.stretch_histogram.figure.title = f"{len(sub_data)} pixels"

        # update the n_bins since this may be a new layer
        self._histogram_nbins_changed()
        # update the curve/colorbar
        self._update_stretch_curve(msg)

    @observe('image_color_mode_value', 'image_color_value', 'image_colormap_value',
             'image_contrast_value', 'image_bias_value',
             'stretch_hist_nbins',
             'stretch_curve_visible',
             'stretch_function_value', 'stretch_vmin_value', 'stretch_vmax_value',
             'stretch_params_value', 'stretch_preset_value',
             'layer_multiselect'
             )
    @skip_if_no_updates_since_last_active()
    def _update_stretch_curve(self, msg=None):
        if not self._viewer_is_image_viewer() or not hasattr(self, 'stretch_histogram'):
            # don't update histogram if selected viewer is not an image viewer,
            # or the stretch histogram hasn't been initialized:
            return

        if self.layer_multiselect and len(self.layer.selected) > 1:
            # currently only support single-layer, if multiple layers are selected, the plot
            # will be hidden in the UI
            return

        # could be multi or single-viewer and/or multi-layer with a single entry,
        # either way, we act on the first entry
        layer = self.layer.selected_obj[0]
        while isinstance(layer, list):
            if not len(layer):
                return
            layer = layer[0]

        if isinstance(layer.layer, GroupedSubset):
            # don't update histogram for subsets, will be hidden in UI
            return

        # create the new/updated stretch curve following the colormapping
        # procedure in glue's CompositeArray:
        interval = ManualInterval(self.stretch_vmin_value, self.stretch_vmax_value)
        contrast_bias = ContrastBiasStretch(self.image_contrast_value, self.image_bias_value)
        stretch = layer.state.stretch_object
        layer_cmap = layer.state.cmap

        # show the colorbar
        color_mode = self.image_color_mode_value

        # NOTE: Index 0 in marks is assumed to be the bin centers.
        x = self.stretch_histogram.figure.marks[0].x
        y = np.ones_like(x)

        # Copied from the __call__ internals of glue/viewers/image/composite_array.py
        data = interval(x)
        data = contrast_bias(data, out=data)
        data = stretch(data, out=data)

        if color_mode == 'Colormaps':
            cmap = glue_colormaps[self.image_colormap.text]
            if hasattr(cmap, "get_bad"):
                bad_color = cmap.get_bad().tolist()[:3]
                layer_cmap = cmap.with_extremes(bad=bad_color + [self.image_opacity_value])
            else:
                layer_cmap = cmap

            # Compute colormapped image
            plane = layer_cmap(data)

        else:  # Color (Monochromatic)
            if self.image_color_value is None:
                # do not crash if image_color_value is not yet assigned,
                # _update_stretch_curve observes image_color_value so will get called again
                return False
            # Get color
            color = COLOR_CONVERTER.to_rgba_array(self.image_color_value)[0]
            plane = data[:, np.newaxis] * color
            plane[:, 3] = 1

        plane = np.clip(plane, 0, 1, out=plane)
        ipycolors = [matplotlib.colors.rgb2hex(p, keep_alpha=False) for p in plane]

        colorbar_mark = self.stretch_histogram.marks['colorbar']
        colorbar_mark.x = x
        colorbar_mark.y = y
        colorbar_mark.colors = ipycolors

        # show "knot" locations if the stretch_function is a spline
        if isinstance(stretch, SplineStretch) and self.stretch_curve_visible:
            knot_mark = self.stretch_histogram.marks['stretch_knots']
            knot_mark.x = (self.stretch_vmin_value +
                           np.asarray(stretch._x) * (self.stretch_vmax_value - self.stretch_vmin_value))  # noqa
            # scale to 0.9 so always falls below colorbar (same as for stretch_curve)
            knot_mark.y = 0.9 * np.asarray(stretch._y)
        else:
            self.stretch_histogram.clear_marks('stretch_knots')

        if self.stretch_curve_visible:
            # create a photoshop style "curve" for the stretch function
            curve_x = np.linspace(self.stretch_vmin_value, self.stretch_vmax_value, 50)
            curve_y = interval(curve_x)
            curve_y = contrast_bias(curve_y)
            curve_y = stretch(curve_y)

            curve_mark = self.stretch_histogram.marks['stretch_curve']
            curve_mark.x = curve_x
            curve_mark.y = 0.9 * curve_y
        else:
            self.stretch_histogram.clear_marks('stretch_curve')

        self.stretch_histogram._refresh_marks()

    @observe('stretch_vmin_value')
    def _stretch_vmin_changed(self, msg=None):
        self.stretch_histogram.marks['vmin'].x = [self.stretch_vmin_value, self.stretch_vmin_value]

    @observe('stretch_vmax_value')
    def _stretch_vmax_changed(self, msg=None):
        self.stretch_histogram.marks['vmax'].x = [self.stretch_vmax_value, self.stretch_vmax_value]

    @observe("stretch_hist_nbins")
    def _histogram_nbins_changed(self, msg={}):
        if self.stretch_histogram is None:
            return
        if self.stretch_hist_nbins == '' or self.stretch_hist_nbins < 1:
            return
        self.stretch_histogram.viewer.state.hist_n_bin = self.stretch_hist_nbins
        # for some reason, this resets the internal marks, so we need to ensure the manual
        # marks are still plotted
        self.stretch_histogram._refresh_marks()

    def set_histogram_limits(self, x_min=None, x_max=None, y_min=None, y_max=None):
        # NOTE: leaving this out of user API until API is finalized with interactive setting
        self.stretch_histogram.set_limits(x_min=x_min, x_max=x_max,
                                          y_min=y_min, y_max=y_max)

    def _viewer_is_image_viewer(self):
        # Import here to prevent circular import (and not at the top of the method so the import
        # check is avoided, whenever possible).
        from jdaviz.configs.imviz.plugins.viewers import ImvizImageView
        from jdaviz.configs.cubeviz.plugins.viewers import CubevizImageView
        from jdaviz.configs.mosviz.plugins.viewers import MosvizImageView, MosvizProfile2DView
        from jdaviz.configs.rampviz.plugins.viewers import RampvizImageView

        def _is_image_viewer(viewer):
            return isinstance(viewer, (ImvizImageView, CubevizImageView,
                                       MosvizImageView, MosvizProfile2DView,
                                       RampvizImageView))

        viewers = self.viewer.selected_obj
        if not isinstance(viewers, list):
            viewers = [viewers]

        return np.all([_is_image_viewer(viewer) for viewer in viewers])

    def image_segmentation_map_presets(self, *args, **kwargs):
        # if 'Random' colormap is used for visualizing image segmentation,
        # ensure the stretch limits are the min and max, the stretch function
        # is linear, the contrast is 1.0, and the bias is 0.5. This ensures
        # that all label colors are unique:
        if self.image_colormap_value != 'Random':
            return

        self.stretch_preset.value = 100
        self.stretch_function.value = 'linear'
        self.image_contrast_value = 1
        self.image_bias_value = 0.5

    def vue_image_segmentation_map_presets(self, *args, **kwargs):
        self.image_segmentation_map_presets(*args, **kwargs)



import os

import gwcs
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import NoConvergence
from echo import delay_callback
from glue.config import colormaps
from glue.core import Data

from jdaviz.utils import get_top_layer_index, get_reference_image_data, data_has_valid_wcs
from jdaviz.core.events import SnackbarMessage, AstrowidgetMarkersChangedMessage

__all__ = ['AstrowidgetsImageViewerMixin']


class AstrowidgetsImageViewerMixin:
    """This class implements ``astrowidgets`` API for Jdaviz *image* viewer.
    This does not provide a fully functional viewer, but rather
    should be used as mixin into an existing viewer subclass.

    A child class that uses this must run :meth:`init_astrowidgets_api`
    within its ``__init__``.

    .. note:: Do not use this for spectral or cube viewers.

    """
    RESERVED_MARKER_SET_NAMES = ['all']

    # __init__ not called, so use this to setup.
    def init_astrowidgets_api(self):
        """This method must be called in child class ``__init__``."""
        # Markers
        self._marktags = set()
        self._default_mark_tag_name = 'default-marker-name'
        # marker shape not settable: https://github.com/glue-viz/glue/issues/2202
        self.marker = {'color': 'red', 'alpha': 1.0, 'markersize': 5}

    def save(self, filename):
        """Save out the current image view to given PNG filename.

        Parameters
        ----------
        filename : str
            PNG filename. If the given file already exists, it will be
            silently overwritten.

        """
        if not filename.lower().endswith('.png'):
            filename = filename + '.png'

        # https://github.com/bqplot/bqplot/pull/1397
        def on_png_received(data):
            with open(os.path.expanduser(filename), 'bw') as f:
                f.write(data)

        self.figure.get_png_data(on_png_received)

    def center_on(self, point):
        """Centers the view on a particular point on the top visible layer.
        The data label of the top visible layer can be queried using the viewer's
        `~jdaviz.configs.imviz.plugins.viewers.ImvizImageView.top_visible_data_label`
        property.

        Parameters
        ----------
        point : tuple or `~astropy.coordinates.SkyCoord`
            If tuple of ``(X, Y)`` is given, it is assumed
            to be in data coordinates and 0-indexed.

        Raises
        ------
        AttributeError
            Sky coordinates are given but image does not have a valid WCS.

        """
        i_top = get_top_layer_index(self)
        if i_top is None:
            return

        image = self.layers[i_top].layer

        if isinstance(point, SkyCoord):
            if data_has_valid_wcs(image):
                try:
                    point = image.coords.world_to_pixel(point)  # 0-indexed X, Y
                except NoConvergence as e:  # pragma: no cover
                    self.session.hub.broadcast(SnackbarMessage(
                        f'{point} is likely out of bounds: {repr(e)}',
                        color="warning", sender=self))
                    return
            else:
                raise AttributeError(f'{getattr(image, "label", None)} does not have a valid WCS')

        if not np.all(np.isfinite(point)):
            return
        elif hasattr(self, '_get_real_xy'):
            # User gives pixel wrt top layer but we want reference data location.
            pix = self._get_real_xy(image, point[0], point[1], reverse=True)
        else:  # pragma: no cover
            pix = point

        width = self.state.x_max - self.state.x_min
        height = self.state.y_max - self.state.y_min
        new_xmin = pix[0] - (width * 0.5)
        new_ymin = pix[1] - (height * 0.5)

        self.set_limits(
            x_min=new_xmin, x_max=new_xmin + width,
            y_min=new_ymin, y_max=new_ymin + height)

    def offset_by(self, dx, dy):
        """Move the center to a point that is given offset
        away from the current center.

        Parameters
        ----------
        dx, dy : float or `~astropy.units.Quantity`
            Offset value. Without a unit, assumed to be pixel offsets.
            If a unit is attached, offset by pixel or sky is assumed from
            the unit.

        Raises
        ------
        AttributeError
            Sky offset is given but image does not have a valid WCS.

        ValueError
            Offsets are of different types.

        astropy.units.core.UnitTypeError
            Sky offset has invalid unit.

        """
        i_top = get_top_layer_index(self)
        image = self.layers[i_top].layer
        width = self.state.x_max - self.state.x_min
        height = self.state.y_max - self.state.y_min
        x_cen = self.state.x_min + (width * 0.5)
        y_cen = self.state.y_min + (height * 0.5)
        if hasattr(self, '_get_real_xy'):
            real_cen = self._get_real_xy(image, x_cen, y_cen)
        else:
            real_cen = (x_cen, y_cen)

        dx, dx_coord = _offset_is_pixel_or_sky(dx)
        dy, dy_coord = _offset_is_pixel_or_sky(dy)

        if dx_coord != dy_coord:
            raise ValueError(f'dx is of type {dx_coord} but dy is of type {dy_coord}')

        if dx_coord == 'wcs':
            if data_has_valid_wcs(image):
                # To avoid distortion headache, assume offset is relative to
                # displayed center.
                sky_cen = image.coords.pixel_to_world(real_cen[0], real_cen[1])
                new_cen = sky_cen.spherical_offsets_by(dx, dy)
            else:
                raise AttributeError(f'{getattr(image, "label", None)} does not have a valid WCS')
        else:
            new_cen = (real_cen[0] + dx, real_cen[1] + dy)

        self.center_on(new_cen)

    @property
    def zoom_level(self):
        """
        The zoom level for an image viewer (not linked by WCS).

        .. warning::
            When a viewer is linked by WCS, the result corresponds
            to the ``zoom_level`` of the reference data.

        * 1 means real-pixel-size.
        * 2 means zoomed in by a factor of 2.
        * 0.5 means zoomed out by a factor of 2.
        * 'fit' means zoomed to fit the whole image width into display.

        """
        if self.shape is None:  # pragma: no cover
            raise ValueError('Viewer is still loading, try again later')

        if hasattr(self, '_get_real_xy'):
            image, i_ref = get_reference_image_data(self.jdaviz_app, self.reference)
            # TODO: Do we want top layer instead?
            # i_top = get_top_layer_index(self)
            # image = self.layers[i_top].layer
            real_min = self._get_real_xy(image, self.state.x_min, self.state.y_min)
            real_max = self._get_real_xy(image, self.state.x_max, self.state.y_max)
        else:
            real_min = (self.state.x_min, self.state.y_min)
            real_max = (self.state.x_max, self.state.y_max)
        screenx = self.shape[1]
        screeny = self.shape[0]
        zoom_x = screenx / abs(real_max[0] - real_min[0])
        zoom_y = screeny / abs(real_max[1] - real_min[1])

        return max(zoom_x, zoom_y)  # Similar to Ginga get_scale()

    # Loosely based on glue/viewers/image/state.py
    @zoom_level.setter
    def zoom_level(self, val):
        if ((not isinstance(val, (int, float)) and val != 'fit') or
                (isinstance(val, (int, float)) and val <= 0)):
            raise ValueError(f'Unsupported zoom level: {val}')

        if (self.shape is None or
                self.state.x_att is None or self.state.y_att is None):  # pragma: no cover
            return

        # Zoom on X and Y will auto-adjust.
        if val == 'fit':
            self.state.reset_limits()
            return

        new_dx = self.shape[1] * 0.5 / val
        if hasattr(self, '_get_real_xy'):
            image, i_ref = get_reference_image_data(self.jdaviz_app, self.reference)
            # TODO: Do we want top layer instead?
            # i_top = get_top_layer_index(self)
            # image = self.layers[i_top].layer
            real_min = self._get_real_xy(image, self.state.x_min, self.state.y_min)
            real_max = self._get_real_xy(image, self.state.x_max, self.state.y_max)
            cur_xcen = (real_min[0] + real_max[0]) * 0.5
            new_x_min = self._get_real_xy(image, cur_xcen - new_dx - 0.5, real_min[1], reverse=True)[0]  # noqa: E501
            new_x_max = self._get_real_xy(image, cur_xcen + new_dx - 0.5, real_max[1], reverse=True)[0]  # noqa: E501
        else:
            cur_xcen = (self.state.x_min + self.state.x_max) * 0.5
            new_x_min = cur_xcen - new_dx - 0.5
            new_x_max = cur_xcen + new_dx - 0.5

        with delay_callback(self.state, 'x_min', 'x_max'):
            self.state.x_min = new_x_min
            self.state.x_max = new_x_max

        # We need to adjust the limits in here to avoid triggering all
        # the update events then changing the limits again.
        self.state._adjust_limits_aspect()

    # Discussion on why we need two different ways to set zoom at
    # https://github.com/astropy/astrowidgets/issues/144
    def zoom(self, val):
        """Zoom in or out by the given factor.

        .. warning::
            When a viewer is linked by WCS, this method accidentally
            changes the center also; see the warning in ``zoom_level``.

        Parameters
        ----------
        val : int or float
            The zoom level to zoom the image.
            See `zoom_level`.

        Raises
        ------
        ValueError
            Invalid zoom factor.

        """
        if not isinstance(val, (int, float)):
            raise ValueError(f"zoom only accepts int or float but got '{val}'")
        self.zoom_level = self.zoom_level * val

    @property
    def colormap_options(self):
        """List of colormap names."""
        # Do not sort to match Plot Options order.
        return [member[0] for member in colormaps.members]

    def set_colormap(self, cmap):
        """Set colormap to the given colormap name.

        Parameters
        ----------
        cmap : str
            Colormap name. Possible values can be obtained from
            :meth:`colormap_options`.

        Raises
        ------
        ValueError
            Invalid colormap name.

        """
        cm = None
        for member in colormaps.members:
            if member[0] == cmap:
                cm = member[1]
                break

        if cm is None:
            raise ValueError(f"Invalid colormap '{cmap}', must be one of {self.colormap_options}")

        i_top = get_top_layer_index(self)
        self.state.layers[i_top].cmap = cm

    @property
    def stretch_options(self):
        """List of all available options for image stretching.

        Their ``astropy.visualization`` counterparts are also accepted, as follows:

        * ``'arcsinh'``: ``astropy.visualization.AsinhStretch``
        * ``'linear'``: ``astropy.visualization.LinearStretch``
        * ``'log'``: ``astropy.visualization.LogStretch``
        * ``'sqrt'``: ``astropy.visualization.SqrtStretch``

        """
        # TODO: Is there a better way to access this in Glue? See glue/viewers/image/state.py
        return ['arcsinh', 'linear', 'log', 'sqrt']

    @property
    def stretch(self):
        """The image stretching algorithm in use."""
        i_top = get_top_layer_index(self)
        return self.state.layers[i_top].stretch

    @stretch.setter
    def stretch(self, val):
        valid_vals = self.stretch_options

        if isinstance(val, type):  # is a class
            # Translate astropy.visualization
            from astropy.visualization import AsinhStretch, LinearStretch, LogStretch, SqrtStretch
            if issubclass(val, AsinhStretch):
                val = 'arcsinh'
            elif issubclass(val, LinearStretch):
                val = 'linear'
            elif issubclass(val, LogStretch):
                val = 'log'
            elif issubclass(val, SqrtStretch):
                val = 'sqrt'
            else:
                raise ValueError(f"Invalid stretch {val}, must be one of {valid_vals}")
        elif val not in valid_vals:
            raise ValueError(f"Invalid stretch '{val}', must be one of {valid_vals}")

        i_top = get_top_layer_index(self)
        self.state.layers[i_top].stretch = val

    @property
    def autocut_options(self):
        """List of all available options for automatic image cut levels."""
        # See glue-jupyter/bqplot/image/state.py#L29
        return ['minmax', '99.5%', '99%', '95%', '90%']

    @property
    def cuts(self):
        """Current image cut levels.

        To set new cut levels, either provide a tuple of ``(low, high)`` values
        or one of the options from `autocut_options`.

        """
        i_top = get_top_layer_index(self)
        return self.state.layers[i_top].v_min, self.state.layers[i_top].v_max

    # TODO: Support astropy.visualization, see https://github.com/glue-viz/glue/issues/2218
    @cuts.setter
    def cuts(self, val):
        i_top = get_top_layer_index(self)

        if isinstance(val, str):  # autocut
            if val == 'minmax':
                val = 100
            elif val == '99.5%':
                val = 99.5
            elif val == '99%':
                val = 99
            elif val == '95%':
                val = 95
            elif val == '90%':
                val = 90
            else:
                raise ValueError(f"Invalid autocut '{val}', must be one of {self.autocut_options}")
            self.state.layers[i_top].percentile = val
        else:  # (low, high)
            if (not isinstance(val, (list, tuple)) or len(val) != 2
                    or not np.all([isinstance(x, (int, float)) for x in val])):
                raise ValueError(f"Invalid cut levels {val}, must be (low, high)")
            self.state.layers[i_top].v_min = val[0]
            self.state.layers[i_top].v_max = val[1]

    @property
    def marker(self):
        """Marker to use.

        Marker can be set as follows; e.g.::

            {'color': 'red', 'alpha': 1.0, 'markersize': 3}
            {'color': '#ff0000', 'alpha': 0.5, 'markersize': 10}
            {'color': (1, 0, 0)}

        The valid properties for Glue markers are listed at
        https://glue-core.readthedocs.io/en/latest/api/glue.core.visual.VisualAttributes.html

        """
        return self._marker_dict

    @marker.setter
    def marker(self, val):
        # Validation: Ideally Glue should do this but we have to due to
        # https://github.com/glue-viz/glue/issues/2203
        given = set(val.keys())
        allowed = set(('color', 'alpha', 'markersize', 'fill'))
        if not given.issubset(allowed):
            raise KeyError(f'Invalid attribute(s): {given - allowed}')
        if 'color' in val:
            from matplotlib.colors import ColorConverter
            ColorConverter().to_rgb(val['color'])  # ValueError: Invalid RGBA argument
        if 'alpha' in val:
            alpha = val['alpha']
            if not isinstance(alpha, (int, float)) or alpha < 0 or alpha > 1:
                raise ValueError(f'Invalid alpha: {alpha}')
        if 'markersize' in val:
            size = val['markersize']
            if not isinstance(size, (int, float)):
                raise ValueError(f'Invalid marker size: {size}')
        if 'fill' in val:
            fill = val['fill']
            if not isinstance(fill, bool):
                raise ValueError(f'Invalid fill: {fill}')

        # Only set this once we have successfully validated a marker.
        # Those not set here use Glue defaults.
        self._marker_dict = val

    def _validate_marker_name(self, marker_name):
        """Raise an error if the marker_name is not allowed."""
        if marker_name in self.RESERVED_MARKER_SET_NAMES:
            raise ValueError(
                f"The marker name {marker_name} is not allowed. Any name is "
                f"allowed except these: {', '.join(self.RESERVED_MARKER_SET_NAMES)}")

    def add_markers(self, table, x_colname='x', y_colname='y',
                    skycoord_colname='coord', use_skycoord=False,
                    marker_name=None):
        """Creates markers w.r.t. the reference image at given points
        in the table.

        .. note:: Use `marker` to change marker appearance.

        .. note::
           Once markers are added, linking cannot be changed.  To change linking options,
           remove and re-add the markers manually.

        Parameters
        ----------
        table : `~astropy.table.Table`
            Table containing marker locations.

        x_colname, y_colname : str
            Column names for X and Y.
            Coordinates must be 0-indexed.

        skycoord_colname : str
            Column name with `~astropy.coordinates.SkyCoord` objects.

        use_skycoord : bool
            If `True`, use ``skycoord_colname`` to mark.
            Otherwise, use ``x_colname`` and ``y_colname``.

        marker_name : str, optional
            Name to assign the markers in the table. Providing a name
            allows markers to be removed by name at a later time.

        Raises
        ------
        AttributeError
            Sky coordinates are given but reference image does not have a valid WCS.

        ValueError
            Invalid marker name.

        """
        from glue.viewers.scatter.state import ScatterLayerState

        if marker_name is None:
            marker_name = self._default_mark_tag_name

        self._validate_marker_name(marker_name)
        jglue = self.session.application

        # Link markers to top visible image data or reference data.
        if not use_skycoord and hasattr(self, '_get_real_xy'):
            i_top = get_top_layer_index(self)
            image = self.layers[i_top].layer
        else:
            image = self.state.reference_data

        # TODO: Is Glue smart enough to no-op if link already there?
        if use_skycoord:
            if not data_has_valid_wcs(image):
                raise AttributeError(f'{getattr(image, "label", None)} does not have a valid WCS')
            sky = table[skycoord_colname]
            t_glue = Data(marker_name, ra=sky.ra.deg, dec=sky.dec.deg)
            dcomps = image.components
            if (isinstance(image.coords, gwcs.WCS) and
                    image.coords.output_frame.reference_frame.name != 'galactic' and
                    'Lon' in dcomps and 'Lat' in dcomps):
                ra_str = 'Lon'
                dec_str = 'Lat'
            else:
                ra_str = 'Right Ascension'
                dec_str = 'Declination'
            with jglue.data_collection.delay_link_manager_update():
                jglue.data_collection[marker_name] = t_glue
                jglue.add_link(t_glue, 'ra', image, ra_str)
                jglue.add_link(t_glue, 'dec', image, dec_str)
        else:
            t_glue = Data(marker_name, **table[x_colname, y_colname])
            with jglue.data_collection.delay_link_manager_update():
                jglue.data_collection[marker_name] = t_glue
                jglue.add_link(t_glue, x_colname, image, image.pixel_component_ids[1].label)
                jglue.add_link(t_glue, y_colname, image, image.pixel_component_ids[0].label)

        try:
            self.add_data(t_glue)
        except Exception as e:  # pragma: no cover
            self.session.hub.broadcast(SnackbarMessage(
                f"Failed to add markers '{marker_name}': {repr(e)}",
                color="warning", sender=self))
        else:
            # Only can set alpha and color using self.add_data(), so brute force here instead.
            # https://github.com/glue-viz/glue/issues/2201
            for lyr in self.state.layers:
                if isinstance(lyr, ScatterLayerState) and lyr.layer.label == marker_name:
                    for key, val in self.marker.items():
                        setattr(lyr, {'markersize': 'size'}.get(key, key), val)
                    break

            self.jdaviz_app.set_data_visibility(self.reference_id, marker_name,
                                                visible=True, replace=False)

            self._marktags.add(marker_name)

            self.session.hub.broadcast(AstrowidgetMarkersChangedMessage(True, sender=self))

    def remove_markers(self, marker_name=None):
        """Remove some but not all of the markers by name used when
        adding the markers.

        Parameters
        ----------
        marker_name : str
            Name used when the markers were added.
            If not given, will delete markers added under default name.

        """
        if marker_name is None:
            marker_name = self._default_mark_tag_name

        # TODO: How to test manually created tiled viewers in CI?
        if marker_name not in self._marktags:  # pragma: no cover
            self.session.hub.broadcast(SnackbarMessage(
                f"Failed to remove markers '{marker_name}': Not added by this viewer",
                color="warning", sender=self))
            return

        try:
            i = self.session.application.data_collection.labels.index(marker_name)
        except ValueError as e:  # pragma: no cover
            self.session.hub.broadcast(SnackbarMessage(
                f"Failed to remove markers '{marker_name}': {repr(e)}",
                color="warning", sender=self))
            return

        data = self.session.application.data_collection[i]
        self.session.application.data_collection.remove(data)
        self._marktags.remove(marker_name)

        self.session.hub.broadcast(AstrowidgetMarkersChangedMessage(len(self._marktags) > 0,
                                                                    sender=self))

    def reset_markers(self):
        """Delete all markers."""
        # Grab the entire list of marker names before iterating
        # otherwise what we are iterating over changes.
        for marker_name in list(self._marktags):
            self.remove_markers(marker_name=marker_name)


def _offset_is_pixel_or_sky(x):
    if isinstance(x, u.Quantity):
        if x.unit in (u.dimensionless_unscaled, u.pix):
            coord = 'data'
            val = x.value
        else:
            coord = 'wcs'
            val = x  # Can stay Quantity
    else:
        coord = 'data'
        val = x

    return val, coord



import numpy as np
from astropy import units as u
from astropy.table import QTable, Table as AstropyTable
from astropy.coordinates import SkyCoord
from echo import delay_callback
from traitlets import List, Unicode, Bool, Int, observe

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, ViewerSelectMixin,
                                        FileImportSelectPluginComponent, HasFileImportSelect,
                                        with_spinner)
from jdaviz.core.custom_traitlets import IntHandleEmpty
from jdaviz.core.events import CatalogResultsChangedMessage, CatalogSelectClickEventMessage
from jdaviz.core.marks import CatalogMark
from jdaviz.core.template_mixin import Table, TableMixin
from jdaviz.core.user_api import PluginUserApi
from jdaviz.utils import get_top_layer_index

__all__ = ['Catalogs']


@tray_registry('imviz-catalogs', label="Catalog Search")
class Catalogs(PluginTemplateMixin, ViewerSelectMixin, HasFileImportSelect, TableMixin):
    """
    See the :ref:`Catalog Search Plugin Documentation <imviz-catalogs>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :meth:`zoom_to_selected`
    """
    template_file = __file__, "catalogs.vue"
    uses_active_status = Bool(True).tag(sync=True)
    catalog_items = List([]).tag(sync=True)
    catalog_selected = Unicode("").tag(sync=True)
    results_available = Bool(False).tag(sync=True)
    number_of_results = Int(0).tag(sync=True)
    max_sources = IntHandleEmpty(1000).tag(sync=True)

    # setting the default table headers and values
    _default_table_values = {
            'Right Ascension (degrees)': np.nan,
            'Declination (degrees)': np.nan,
            'Object ID': np.nan,
            'id': np.nan,
            'x_coord': np.nan,
            'y_coord': np.nan
            }

    headers = ['Right Ascension (degrees)', 'Declination (degrees)',
               'Object ID', 'x_coord', 'y_coord']

    table_selected_widget = Unicode().tag(sync=True)

    @property
    def user_api(self):
        return PluginUserApi(self, expose=('clear_table', 'export_table', 'import_catalog',
                                           'zoom_to_selected', 'select_rows',
                                           'select_all', 'select_none'))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Query catalog for objects within region on sky.'

        cat_options = ['SDSS', 'Gaia']
        if not self.app.state.settings.get('server_is_remote', False):
            cat_options.append('From File...')
        self.catalog = FileImportSelectPluginComponent(self,
                                                       items='catalog_items',
                                                       selected='catalog_selected',
                                                       manual_options=cat_options)

        # set the custom file parser for importing catalogs
        self.catalog._file_parser = self._file_parser
        self._marker_name = 'catalog_results'

        # initializing the headers in the table that is displayed in the UI
        self.table.headers_avail = self.headers
        self.table.headers_visible = self.headers
        self.table._default_values_by_colname = self._default_table_values
        self.table._selected_rows_changed_callback = self._table_selection_changed
        # note: `item_key` is the name of a column in the
        # table that has unique values for in row
        self.table.item_key = 'id'
        self.table.show_rowselect = True

        def clear_table_callback():
            # gets the current viewer
            viewer = self.viewer.selected_obj

            # resetting values
            self.results_available = False
            self.number_of_results = 0

            if self._marker_name in self.app.data_collection.labels:
                # all markers are removed from the viewer
                viewer.remove_markers(marker_name=self._marker_name)

        self.table._clear_callback = clear_table_callback

        self.table_selected = Table(self, name='table_selected')
        self.table_selected.clear_btn_lbl = 'Clear Selection'
        self.table_selected.show_if_empty = False

        def clear_selected_table_callback():
            self.table.select_none()

        self.table_selected._clear_callback = clear_selected_table_callback
        self.table_selected_widget = 'IPY_MODEL_'+self.table_selected.model_id

        self.docs_description = "Queries an area encompassed by the viewer using\
                                 a specified catalog and marks all the objects\
                                 found within the area."

        self.session.hub.subscribe(self, CatalogSelectClickEventMessage,
                                   self._on_catalog_select_click_event)

    @staticmethod
    def _file_parser(path):
        if isinstance(path, AstropyTable):  # includes QTable
            from_file_string = f'API: {path.__class__.__name__} object'
            return '', {from_file_string: path}

        try:
            table = QTable.read(path)
        except Exception:
            return 'Could not parse file with astropy.table.QTable.read', {}

        if not table.colnames:  # Ensure the file has columns
            return "File contains no columns", {}

        if 'sky_centroid' not in table.colnames:
            return 'Table does not contain required sky_centroid column', {}

        return '', {path: table}

    def _on_catalog_select_click_event(self, msg):
        xs, ys = self.table._qtable['x_coord'], self.table._qtable['y_coord']
        # nearest point
        distsq = (xs - msg.x)**2 + (ys - msg.y)**2
        ind = np.argmin(distsq)
        item = self.table.items[ind]
        if item in self.table.selected_rows:
            self.table.selected_rows = [sr for sr in self.table.selected_rows if sr != item]
        else:
            self.table.selected_rows += [item]
        self._table_selection_changed()

    @with_spinner()
    def search(self, error_on_fail=False):
        """Search the catalog, display markers on the viewer, and return results if available.

        Parameters
        ----------
        error_on_fail : bool
            Throw exception when query fails instead of a red snackbar message.
            This is useful for debugging.

        Returns
        -------
        skycoords : `~astropy.coordinates.SkyCoord` or `None`
            Sky coordinates (or None if no results available).

        """
        # calling clear in the case the user forgot after searching
        self.clear_table()

        # gets the current viewer
        viewer = self.viewer.selected_obj

        # nothing happens in the case there is no image in the viewer
        # additionally if the data does not have WCS
        if viewer.state.reference_data is None or viewer.state.reference_data.coords is None:
            self.results_available = False
            self.number_of_results = 0
            return
        # obtains the center point of the current image and converts the point into sky coordinates
        x_center = (viewer.state.x_min + viewer.state.x_max) * 0.5
        y_center = (viewer.state.y_min + viewer.state.y_max) * 0.5
        skycoord_center = viewer.state.reference_data.coords.pixel_to_world(x_center, y_center)

        # obtains the viewer's zoom limits based on the visible layer
        ny, nx = viewer.state.reference_data.shape
        zoom_x_min = max(0, viewer.state.x_min)
        zoom_x_max = min(nx, viewer.state.x_max)
        zoom_y_min = max(0, viewer.state.y_min)
        zoom_y_max = min(ny, viewer.state.y_max)
        zoom_coordinate = viewer.state.reference_data.coords.pixel_to_world(
            [zoom_x_min, zoom_x_min, zoom_x_max, zoom_x_max],
            [zoom_y_min, zoom_y_max, zoom_y_max, zoom_y_min])

        # radius for querying the region is based on:
        # the distance between the longest zoom limits and the center point
        zoom_radius = max(skycoord_center.separation(zoom_coordinate))

        max_sources_used = False

        # conducts search based on SDSS
        if self.catalog_selected == "SDSS":
            from astroquery.sdss import SDSS
            r_max = 3 * u.arcmin
            src_id_colname = "objid"

            # queries the region (based on the provided center point and radius)
            # finds all the sources in that region
            try:
                if zoom_radius > r_max:  # SDSS now has radius max limit
                    self.hub.broadcast(SnackbarMessage(
                        f"Radius for {self.catalog_selected} has max radius of {r_max} but got "
                        f"{zoom_radius.to(u.arcmin)}, using {r_max}.",
                        color='warning', sender=self))
                    zoom_radius = r_max
                query_region_result = SDSS.query_region(skycoord_center, radius=zoom_radius,
                                                        data_release=17)
                if len(query_region_result) > self.max_sources:
                    query_region_result = query_region_result[:self.max_sources]
                    max_sources_used = True
            except Exception as e:  # nosec
                errmsg = (f"Failed to query {self.catalog_selected} with c={skycoord_center} and "
                          f"r={zoom_radius}: {repr(e)}")
                if error_on_fail:
                    raise Exception(errmsg) from e
                self.hub.broadcast(SnackbarMessage(errmsg, color='error', sender=self))
                query_region_result = None

            if query_region_result is None:
                self.results_available = True
                self.number_of_results = 0
                self.app._catalog_source_table = None
                viewer.remove_markers(marker_name=self._marker_name)
                return

            # TODO: Filter this table the same way as the actual displayed markers.
            # attach the table to the app for Python extraction
            self.app._catalog_source_table = query_region_result

        elif self.catalog_selected == 'Gaia':
            from astroquery.gaia import Gaia

            Gaia.ROW_LIMIT = self.max_sources
            sources = Gaia.query_object(skycoord_center, radius=zoom_radius,
                                        columns=('source_id', 'ra', 'dec')
                                        )
            if "SOURCE_ID" in sources.colnames:  # Case could flip non-deterministically
                src_id_colname = "SOURCE_ID"
            else:
                src_id_colname = "source_id"
            if len(sources) == self.max_sources:
                max_sources_used = True
            self.app._catalog_source_table = sources

        elif self.catalog_selected == 'From File...':
            # all exceptions when going through the UI should have prevented setting this path
            # but this exceptions might be raised here if setting from_file from the UI
            table = self.catalog.selected_obj
            column_names = table.colnames
            self.table.headers_avail = self.headers + [
                col for col in column_names if col not in self.headers]
            self.table.headers_visible = self.headers
            if len(table['sky_centroid']) > self.max_sources:
                self.app._catalog_source_table = table[:self.max_sources]
                max_sources_used = True
            else:
                self.app._catalog_source_table = table
        else:
            self.results_available = False
            self.number_of_results = 0
            self.app._catalog_source_table = None
            raise NotImplementedError(f"{self.catalog_selected} not a supported catalog")

        self.results_available = True
        if not len(self.app._catalog_source_table):
            self.number_of_results = 0
            self.app._catalog_source_table = None
            return

        if max_sources_used:
            snackbar_message = SnackbarMessage(
                    f"{self.catalog_selected} queried, results returned were limited using max_sources = {self.max_sources}.",  # noqa
                    color="success",
                    sender=self)
            self.hub.broadcast(snackbar_message)

        # Convert to pixel coordinates and filter results to be within viewer bounds
        if self.catalog_selected in ["SDSS", "Gaia"]:
            skycoords = SkyCoord(self.app._catalog_source_table['ra'],
                                 self.app._catalog_source_table['dec'],
                                 unit='deg')
        elif self.catalog_selected in ["From File..."]:
            skycoords = self.app._catalog_source_table['sky_centroid']

        pixel_table = viewer.state.reference_data.coords.world_to_pixel(skycoords)
        self.app._catalog_source_table['x_coord'] = pixel_table[0]
        self.app._catalog_source_table['y_coord'] = pixel_table[1]
        x_coordinates = []
        y_coordinates = []
        source_table = self.app._catalog_source_table
        mask = ((source_table['x_coord'] < zoom_x_min) |
                (source_table['x_coord'] > zoom_x_max) |
                (source_table['y_coord'] < zoom_y_min) |
                (source_table['y_coord'] > zoom_y_max))
        self.app._catalog_source_table = self.app._catalog_source_table[~mask]
        skycoords = skycoords[~mask]

        if self.catalog_selected in ["SDSS", "Gaia"]:
            for row in self.app._catalog_source_table:
                x_coordinates.append(row['x_coord'])
                y_coordinates.append(row['y_coord'])
                row_id = row[src_id_colname]
                # Check if the row contains the required keys
                row_info = {'Right Ascension (degrees)': row['ra'],
                            'Declination (degrees)': row['dec'],
                            'Object ID': row_id.astype(str),
                            'id': len(self.table),
                            'x_coord': row['x_coord'],
                            'y_coord': row['y_coord']}
                self.table.add_item(row_info)

        # NOTE: If performance becomes a problem, see
        # https://docs.astropy.org/en/stable/table/index.html#performance-tips
        elif self.catalog_selected in ["From File..."]:
            for row in self.app._catalog_source_table:  # noqa:E501
                row_info = {
                    'Right Ascension (degrees)': row['sky_centroid'].ra.deg,
                    'Declination (degrees)': row['sky_centroid'].dec.deg,
                    'Object ID': str(row.get('label', f"{len(self.table) + 1}")),
                    'id': len(self.table),
                    'x_coord': row['x_coord'],
                    'y_coord': row['y_coord'],
                }
                x_coordinates.append(row['x_coord'])
                y_coordinates.append(row['y_coord'])
                # Add sky_centroid and label explicitly to row_info
                row_info['sky_centroid'] = row['sky_centroid']
                row_info['label'] = row_info['Object ID']
                for col in table.colnames:
                    if col not in self.headers:  # Skip already processed columns
                        row_info[col] = row[col]

                self.table.add_item(row_info)

        filtered_skycoords = viewer.state.reference_data.coords.pixel_to_world(x_coordinates,
                                                                               y_coordinates)

        # QTable stores all the filtered sky coordinate points to be marked
        catalog_results = QTable({'coord': filtered_skycoords})

        self.number_of_results = len(catalog_results)
        # markers are added to the viewer based on the table
        viewer.marker = {'color': 'blue', 'alpha': 0.8, 'markersize': 30, 'fill': False}
        viewer.add_markers(table=catalog_results, use_skycoord=True, marker_name=self._marker_name)

        msg = CatalogResultsChangedMessage(sender=self)
        self.session.hub.broadcast(msg)

        return skycoords

    def _get_mark(self, viewer):
        matches = [mark for mark in viewer.figure.marks if isinstance(mark, CatalogMark)]
        if len(matches):
            return matches[0]
        mark = CatalogMark(viewer)
        viewer.figure.marks = viewer.figure.marks + [mark]
        return mark

    @property
    def marks(self):
        return {viewer_id: self._get_mark(viewer)
                for viewer_id, viewer in self.app._viewer_store.items()
                if hasattr(viewer, 'figure')}

    @observe('is_active')
    def _on_is_active_changed(self, *args):
        if self.disabled_msg:
            return

        for mark in self.marks.values():
            mark.visible = self.is_active

    def _table_selection_changed(self, msg={}):
        selected_rows = self.table.selected_rows

        self.table_selected._clear_table()
        for selected_row in selected_rows:
            self.table_selected.add_item(selected_row)

        x = [float(coord['x_coord']) for coord in selected_rows]
        y = [float(coord['y_coord']) for coord in selected_rows]
        self._get_mark(self.viewer.selected_obj).update_xy(getattr(x, 'value', x),
                                                           getattr(y, 'value', y))

    def vue_zoom_in(self, *args, **kwargs):
        self.zoom_to_selected()

    def zoom_to_selected(self, padding=0.02, return_bounding_box=False):
        """
        Zoom on the default viewer to a region containing the currently selected
        points in the catalog.

        Parameters
        ----------
        padding : float, optional
            A fractional value representing the padding around the bounding box
            of the selected points. It is applied as a proportion of the largest
            dimension of the current extent of loaded data. Defaults to 0.02.
        return_bounding_box : bool, optional
            If True, returns the bounding box of the zoomed region as
            ((x_min, x_max), (y_min, y_max)). Defaults to False.

        Returns
        -------
        bb : list of float or None
            If there are activley selected rows, and ``return_bounding_box`` is
            True, returns a list containing the bounding
            box coordinates: ``[x_min, x_max, y_min, y_max]``.
            Otherwise, returns `None`.

        """
        viewer = self.viewer.selected_obj  # gets the current viewer

        selected_rows = self.table.selected_rows
        if not selected_rows:  # Check if no rows are selected
            return

        if padding <= 0 or padding > 1:
            raise ValueError("padding must be between 0 (exclusive) and 1 (inclusive).")

        i_top = get_top_layer_index(viewer)
        image = viewer.layers[i_top].layer
        x_min = 99999
        x_max = -99999
        y_min = 99999
        y_max = -99999
        for coord in selected_rows:  # list of dict
            cur_x, cur_y = viewer._get_real_xy(
                image, float(coord['x_coord']), float(coord['y_coord']))[:2]
            if cur_x < x_min:
                x_min = cur_x
            if cur_x > x_max:
                x_max = cur_x
            if cur_y < y_min:
                y_min = cur_y
            if cur_y > y_max:
                y_max = cur_y

        if x_min == x_max and y_min == y_max:  # Only one selected
            pass
        elif x_min >= x_max or y_min >= y_max:
            raise ValueError(
                f"Zoom failed: x_min={x_min}, x_max={x_max}, y_min={y_min}, y_max={y_max}")

        pix_pad = padding * max(x_max, y_max)
        x_min -= pix_pad
        x_max += pix_pad
        y_min -= pix_pad
        y_max += pix_pad
        new_y_min = viewer._get_real_xy(image, x_min, y_min, reverse=True)[1]
        new_y_max = viewer._get_real_xy(image, x_max, y_max, reverse=True)[1]

        # First, we center using image's coordinates.
        viewer.center_on((0.5 * (x_min + x_max), 0.5 * (y_min + y_max)))

        # Then, we zoom using reference data's coordinates. This is important when WCS linked.
        # We cannot use viewer.zoom_level because it is wonky when WCS linked.
        # Given most displays are wider in X, we make sure Y coordinates all fit first
        # and X will naturally all fit within after aspect ratio is taken into account.
        with delay_callback(viewer.state, 'x_min', 'x_max', 'y_min', 'y_max'):
            viewer.state.y_min = new_y_min
            viewer.state.y_max = new_y_max
        viewer.state._adjust_limits_aspect()

        if return_bounding_box:
            return [viewer.state.x_min, viewer.state.x_max, viewer.state.y_min, viewer.state.y_max]

    def import_catalog(self, catalog):
        """
        Import a catalog from a file path.

        Parameters
        ----------
        catalog : str
          Path to a file that can be parsed by astropy QTable
        """
        if isinstance(catalog, str):
            self.catalog.import_file(catalog)
        elif isinstance(catalog, AstropyTable):  # includes QTable
            self.catalog.import_obj(catalog)
        else:  # pragma: no cover
            raise ValueError("catalog must be a string (file path) or Table object")

    def vue_do_search(self, *args, **kwargs):
        # calls self.search() which handles all of the searching logic
        self.search()



import pytest
from astropy.utils.data import download_file

from jdaviz.core.data_formats import identify_helper


# URIs to example JWST/HST files on MAST, and their corresponding jdaviz helpers.
@pytest.mark.remote_data
@pytest.mark.filterwarnings(r"ignore::astropy.wcs.wcs.FITSFixedWarning")
@pytest.mark.parametrize(("uri, expected_helper"), [
    ('mast:HST/product/jclj01010_drz.fits', 'imviz'),
    ('mast:JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_x1d.fits', 'specviz'),
    ('mast:jwst/product/jw01538-o161_s000000001_nirspec_f290lp-g395h-s1600a1_s2d.fits', 'specviz2d'),  # noqa: E501
    ('mast:JWST/product/jw02727-o002_t062_nircam_clear-f277w_i2d.fits', 'imviz'),
    ('mast:JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_s3d.fits', 'cubeviz')])
def test_auto_config_detection(uri, expected_helper):
    url = f'https://mast.stsci.edu/api/v0.1/Download/file/?uri={uri}'
    fn = download_file(url, timeout=100)
    helper_name, hdul = identify_helper(fn)
    hdul.close()
    assert len(helper_name) == 1 and helper_name[0] == expected_helper



from zipfile import ZipFile

import pytest
from numpy.testing import assert_allclose
from astropy.utils.data import download_file

from jdaviz.utils import PRIHDR_KEY, COMMENTCARD_KEY


@pytest.mark.remote_data
@pytest.mark.parametrize('instrument_arg', ('nirspec', None))
def test_nirspec_parser(mosviz_helper, tmp_path, instrument_arg):
    '''
    Tests loading our default MosvizExample notebook data
    Also tests no instrument keyword fallback, and IntraRow linking
    '''

    test_data = 'https://stsci.box.com/shared/static/ovyxi5eund92yoadvv01mynwt8t5n7jv.zip'
    fn = download_file(test_data, cache=True, timeout=30)
    with ZipFile(fn, 'r') as sample_data_zip:
        sample_data_zip.extractall(tmp_path)

    level3_path = tmp_path / 'mosviz_nirspec_data_0.3' / 'level3'

    data_dir = level3_path

    if instrument_arg:
        mosviz_helper.load_data(directory=data_dir, instrument=instrument_arg)
    else:
        # When no instrument is provided, Mosviz now raises an error
        match = ('Ambiguous MOS Instrument: Only JWST NIRSpec, NIRCam, and NIRISS folder parsing'
                 ' are currently supported')
        with pytest.raises(ValueError, match=match):
            mosviz_helper.load_data(directory=data_dir)

        return

    assert len(mosviz_helper.app.data_collection) == 16

    # MOS Table meta should be empty:
    assert len(mosviz_helper.app.data_collection["MOS Table"].meta) == 0

    # Check that the data was loaded in the same order we expect.
    # MOS table always loads first, followed by 5 sets of 1D Spectra, 2D Spectra, then Images
    #   which results in 1D beginning at index 1, 2D at index 6, and Images at index 11
    assert mosviz_helper.app.data_collection[5].meta['SOURCEID'] == 2315
    for i in range(0, 5):
        # Check 1D spectra
        spec1d = mosviz_helper.app.data_collection[i+1]
        assert spec1d.label == f"1D Spectrum {i}"
        # Check 2D spectra
        spec2d = mosviz_helper.app.data_collection[i+6]
        assert spec2d.label == f"2D Spectrum {i}"
        assert int(spec1d.meta['SOURCEID']) == int(spec2d.meta['SOURCEID'])
        assert int(spec1d.meta['mosviz_row']) == int(spec2d.meta['mosviz_row'])
        # Check images
        assert mosviz_helper.app.data_collection[i+11].label == f"Image {i}"

    # Check for expected metadata values
    for data in mosviz_helper.app.data_collection:
        assert PRIHDR_KEY not in data.meta
        assert 'header' not in data.meta

        if 'IMAGE' in data.label:
            assert data.meta['WCSAXES'] == 2
        elif 'Spectrum' in data.label:
            assert data.meta['TARGNAME'] == 'FOO'

    # Test IntraRow linking:
    # Attempts to add the spectrum of another row into the current row's viewers
    # Currently, intrarow linking is disabled. Attemps to load another spectrum into
    # the current spectrum viewer should result in an error

    # Check to make sure our test case isn't from the same row to avoid false positive
    table = mosviz_helper.app.get_viewer('table-viewer')
    table.select_row(0)
    data_label = "1D Spectrum 4"
    assert mosviz_helper.app.data_collection[data_label].meta['mosviz_row'] != table.current_row

    with pytest.raises(NotImplementedError, match='Intra-row plotting not supported'):
        mosviz_helper.app.add_data_to_viewer(viewer_reference='spectrum-viewer',
                                             data_label=data_label)


@pytest.mark.remote_data
def test_nirspec_level2_parser(mosviz_helper, tmp_path):
    '''
    Tests loading our default MosvizExample notebook data
    Also tests no instrument keyword fallback, and IntraRow linking
    '''

    # nirspec_level2_test_data.zip
    test_data = 'https://stsci.box.com/shared/static/4hgfhul8el3ciu3p59aqb1m0dk3b0v8y.zip'
    fn = download_file(test_data, cache=True, timeout=100)
    with ZipFile(fn, 'r') as sample_data_zip:
        sample_data_zip.extractall(tmp_path)

    level3_path = tmp_path / 'jw02756001001_03103_00003_nrs1'

    data_dir = level3_path
    mosviz_helper.load_data(directory=data_dir, instrument='nirspec')

    assert len(mosviz_helper.app.data_collection) == 75


@pytest.mark.remote_data
def test_niriss_parser(mosviz_helper, tmp_path):
    '''
    Tests loading a NIRISS dataset
    This data set is a shortened version of the ERS program GLASS (Program 1324)
    provided by Camilla Pacifici. This is in-flight, "real" JWST data

    The spectra are jw01324001001_15101_00001_nis
    The direct image is jw01324-o001_t001_niriss_clear-f200w
    Please see JWST naming conventions for the above

    The dataset was uploaded to box by Duy Nguyen
    '''
    # TODO: Change back to smaller number (30?) when ITSD is convinced it is them and not us.
    #       Help desk ticket INC0183598, J. Quick.
    # Download data
    test_data = 'https://stsci.box.com/shared/static/cr14xijcg572dglacochctr1kblsr89a.zip'
    fn = download_file(test_data, cache=True, timeout=100)

    # Extract to a known, temporary folder
    with ZipFile(fn, 'r') as sample_data_zip:
        sample_data_zip.extractall(tmp_path)

    mosviz_helper.load_data(directory=tmp_path, instrument="niriss")
    assert len(mosviz_helper.app.data_collection) == 10

    # The MOS Table should be first in the data collection
    dc_tab = mosviz_helper.app.data_collection[0]
    assert dc_tab.label == "MOS Table"
    assert len(dc_tab.meta) == 0

    # The image should be the first "real data" in the data collection
    dc_1 = mosviz_helper.app.data_collection[1]
    assert dc_1.label == "Image jw01324-o001 F200W"
    assert PRIHDR_KEY not in dc_1.meta
    assert COMMENTCARD_KEY not in dc_1.meta
    assert dc_1.meta['bunit_data'] == 'MJy/sr'  # ASDF metadata

    # We should be centered on the coordinates of the first data point
    imview = mosviz_helper.app.get_viewer(mosviz_helper._default_image_viewer_reference_name)
    x_pixcenter = (imview.state.x_max + imview.state.x_min)/2.0
    y_pixcenter = (imview.state.y_max + imview.state.y_min)/2.0
    viewer_center_coord = imview.layers[0].layer.coords.pixel_to_world(x_pixcenter, y_pixcenter)
    assert_allclose(viewer_center_coord.ra.deg, dc_tab["R.A."][0])
    assert_allclose(viewer_center_coord.dec.deg, dc_tab["Dec."][0])

    # Test all the spectra exist
    for dispersion in ('R', 'C'):
        for sourceid in (243, 249):
            spec2d = mosviz_helper.app.data_collection[
                f"F200W Source {sourceid} spec2d {dispersion}"]
            spec1d = mosviz_helper.app.data_collection[
                f"F200W Source {sourceid} spec1d {dispersion}"]

            # Header should be imported from the spec2d files
            assert PRIHDR_KEY in spec2d.meta
            assert COMMENTCARD_KEY in spec2d.meta

            # Header should NOT be imported from spec1d files:
            assert PRIHDR_KEY not in spec1d.meta
            assert COMMENTCARD_KEY in spec1d.meta
            assert 'header' not in spec1d.meta

            # Other Metadata Checks:
            assert spec1d.meta['FILTER'] == f'GR150{dispersion}'
            assert int(spec1d.meta['mosviz_row']) == int(spec2d.meta['mosviz_row'])

            for spec in (spec1d, spec2d):
                assert spec.meta['SOURCEID'] == sourceid


@pytest.mark.remote_data
def test_nircam_parser(mosviz_helper, tmp_path):
    '''
    Tests loading a NIRCam dataset
    '''

    # Download data (nircam_parser_test_data.zip)
    test_data = 'https://stsci.box.com/shared/static/g3tymg7c1yj67gy0hpfreo80inqdm1ps.zip'
    fn = download_file(test_data, cache=True, timeout=100)

    # Extract to a known, temporary folder
    with ZipFile(fn, 'r') as sample_data_zip:
        sample_data_zip.extractall(tmp_path)

    mosviz_helper.load_data(directory=tmp_path / "trimmed_nircam_data", instrument="nircam")

    # The MOS Table should be first in the data collection
    dc_tab = mosviz_helper.app.data_collection[0]
    assert dc_tab.label == "MOS Table"
    assert len(dc_tab.meta) == 0

    # Check that the correct amount of spectra got loaded in the correct order
    assert len(mosviz_helper.app.data_collection) == 31
    assert mosviz_helper.app.data_collection['MOS Table']['Identifier'][0] == 1112
    assert mosviz_helper.app.data_collection[1].label == 'F322W2 Source 1112 spec2d R'
    assert mosviz_helper.app.data_collection[16].label == 'F322W2 Source 1112 spec1d R'


@pytest.mark.remote_data
def test_missing_srctype(mosviz_helper, tmp_path):
    '''
    Tests that data missing the SRCTYPE keyword raises a warning to the user.

    SRCTYPE is required for Mosviz. We do not want to rely on the JWST x1d parser's
    default behavior of overwriting with "POINT" if it doesn't exist, as all NIRISS data
    should have this populated; missing SRCTYPE indicates something went wrong.

    This dataset was our original simulated NIRISS dataset that is missing SRCTYPE.

    '''

    # Download data (niriss_parser_test_data.zip)
    test_data = 'https://stsci.box.com/shared/static/7ndets0vjjsa97la2hvjvm6xvnu5b594.zip'
    fn = download_file(test_data, cache=True, timeout=30)

    # Extract to a known, temporary folder
    with ZipFile(fn, 'r') as sample_data_zip:
        sample_data_zip.extractall(tmp_path)

    with pytest.raises(KeyError, match=r".*The SRCTYPE keyword.*is not populated.*"):
        mosviz_helper.load_data(directory=(tmp_path / 'NIRISS_for_parser_p0171'),
                                instrument="niriss")



import numpy as np
from contextlib import contextmanager
import sys
import os
import time

try:
    from strauss.sonification import Sonification
    from strauss.sources import Events
    from strauss.score import Score
    from strauss.generator import Spectralizer
except ImportError:
    pass

#  smallest fraction of the max audio amplitude that can be represented by a 16-bit signed integer
MINVOL = 1/(2**15 - 1)


@contextmanager
def suppress_stderr():
    with open(os.devnull, "w") as devnull:
        old_stderr = sys.stderr
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stderr = old_stderr


def sonify_spectrum(spec, duration, overlap=0.05, system='mono', srate=44100, fmin=40, fmax=1300,
                    eln=False):
    notes = [["A2"]]
    score = Score(notes, duration)
    # set up spectralizer generator
    generator = Spectralizer(samprate=srate)

    # Lets pick the mapping frequency range for the spectrum...
    generator.modify_preset({'min_freq': fmin, 'max_freq': fmax,
                             'fit_spec_multiples': False,
                             'interpolation_type': 'preserve_power',
                             'equal_loudness_normalisation': eln})

    data = {'spectrum': [spec], 'pitch': [1]}

    # set range in spectral flux representing the maximum and minimum sound frequency power:
    # 0 (numeric): absolute 0 in flux units, such that any flux above 0 will sound.
    # '100' (string): 100th percentile (i.e. maximum value) in spectral flux.
    lims = {'spectrum': (0, '100')}

    # set up source
    sources = Events(data.keys())
    sources.fromdict(data)
    sources.apply_mapping_functions(map_lims=lims)

    # render and play sonification!
    soni = Sonification(score, sources, generator, system, samprate=srate)
    soni.render()
    soni._make_seamless(overlap)

    return soni.loop_channels['0'].values


class CubeListenerData:
    def __init__(self, cube, wlens, samplerate=44100, duration=1, overlap=0.05, buffsize=1024,
                 bdepth=16, wl_unit=None, audfrqmin=50, audfrqmax=1500, eln=False, vol=None):
        self.siglen = int(samplerate*(duration-overlap))
        self.cube = cube
        self.dur = duration
        self.bdepth = bdepth
        self.srate = samplerate
        self.maxval = pow(2, bdepth-1) - 1
        self.fadedx = 0

        if vol is None:
            self.atten_level = 1
        else:
            self.atten_level = int(np.clip((vol/100)**2, MINVOL, 1))

        self.wl_unit = wl_unit
        self.wlens = wlens

        # control fades
        fade = np.linspace(0, 1, buffsize+1)
        self.ifade = fade[:-1]
        self.ofade = fade[::-1][:-1]

        # mapping frequency limits in Hz
        self.audfrqmin = audfrqmin
        self.audfrqmax = audfrqmax

        # do we normalise for equal loudness?
        self.eln = eln

        self.idx1 = 0
        self.idx2 = 0
        self.cbuff = False
        self.cursig = np.zeros(self.siglen, dtype='int16')
        self.newsig = np.zeros(self.siglen, dtype='int16')

        if self.cursig.nbytes * pow(1024, -3) > 2:
            raise Exception("Cube projected to be > 2Gb!")

        self.sigcube = np.zeros((*self.cube.shape[:2], self.siglen), dtype='int16')

    def set_wl_bounds(self, w1, w2):
        """
        set the wavelength bounds for indexing spectra
        """
        wsrt = np.sort([w1, w2])
        self.wl_bounds = tuple(wsrt)

    def sonify_cube(self):
        """
        Iterate through the cube, convert each spectrum to a signal, and store
        in class attributes
        """
        lo2hi = self.wlens.argsort()[::-1]

        t0 = time.time()
        for i in range(self.cube.shape[0]):
            for j in range(self.cube.shape[1]):
                with suppress_stderr():
                    if self.cube[i, j, lo2hi].any():
                        sig = sonify_spectrum(self.cube[i, j, lo2hi], self.dur,
                                              srate=self.srate,
                                              fmin=self.audfrqmin,
                                              fmax=self.audfrqmax,
                                              eln=self.eln)
                        sig = (sig*self.maxval).astype('int16')
                        self.sigcube[i, j, :] = sig
                    else:
                        continue
        self.cursig[:] = self.sigcube[self.idx1, self.idx2, :]
        self.newsig[:] = self.cursig[:]
        t1 = time.time()
        print(f"Took {t1-t0}s to process {self.cube.shape[0]*self.cube.shape[1]} spaxels")

    def player_callback(self, outdata, frames, time, status):
        cur = self.cursig
        new = self.newsig
        sdx = int(time.outputBufferDacTime*self.srate)
        dxs = np.arange(sdx, sdx+frames).astype(int) % self.sigcube.shape[-1]
        if self.cbuff:
            outdata[:, 0] = (cur[dxs] * self.ofade).astype('int16')
            outdata[:, 0] += (new[dxs] * self.ifade).astype('int16')
            self.cursig[:] = self.newsig[:]
            self.cbuff = False
        else:
            outdata[:, 0] = self.cursig[dxs]
        outdata[:, 0] //= self.atten_level



from PyInstaller.utils.hooks import collect_data_files, collect_dynamic_libs

datas = collect_data_files("debugpy")
# we are picking up debugpy/_vendored/pydevd/pydevd_attach_to_process/attach_linux_amd64.dylib
datas = filter(lambda x: not x[0].endswith('.dylib'), datas)
# binaries = collect_dynamic_libs('omp')

# breakpoint()


from PyInstaller.utils.hooks import collect_data_files, collect_dynamic_libs

datas = collect_data_files("skimage", includes=["*.pyi"])
# osx does not like the .dylib directory with signing
# [('.../site-packages/skimage/.dylibs/libomp.dylib', 'skimage/.dylibs')]
binaries = collect_dynamic_libs('skimage')
if binaries and binaries[0][0].endswith('.dylib'):
    assert len(binaries) == 1
    assert binaries[0][0].endswith('.dylibs/libomp.dylib')
    binaries = [
        (binaries[0][0], 'skimage'),
    ]



import numpy as np
import pytest


# Some API might be going through deprecation, so ignore the warning.
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_plugin_user_apis(imviz_helper):
    for plugin_name, plugin_api in imviz_helper.plugins.items():
        plugin = plugin_api._obj
        for attr in plugin_api._expose:
            assert hasattr(plugin, attr)


def test_create_new_viewer(imviz_helper, image_2d_wcs):
    # starts with one (default) viewer
    assert len(imviz_helper.app.get_viewer_ids()) == 1
    arr = np.ones((10, 10))

    data_label = 'image-data'
    viewer_name = 'user-created-viewer'
    imviz_helper.load_data(arr, data_label=data_label, show_in_viewer=False)
    imviz_helper.create_image_viewer(viewer_name=viewer_name)

    returned_data = imviz_helper.get_data(data_label)
    assert len(returned_data.shape) == 2

    # new image viewer created
    assert len(imviz_helper.app.get_viewer_ids()) == 2

    # there should be no data in the new viewer
    assert len(imviz_helper.app.get_viewer(viewer_name).data()) == 0

    # then add data, and check that data were added to the new viewer
    imviz_helper.app.add_data_to_viewer(viewer_name, data_label)
    assert len(imviz_helper.app.get_viewer(viewer_name).data()) == 1

    # remove data from the new viewer, check that it was removed
    imviz_helper.app.remove_data_from_viewer(viewer_name, data_label)
    assert len(imviz_helper.app.get_viewer(viewer_name).data()) == 0



import numpy as np
import astropy.units as u
from functools import cached_property

from jdaviz.core.marks import SliceIndicatorMarks

__all__ = ['WithSliceIndicator', 'WithSliceSelection']


class WithSliceIndicator:
    @property
    def slice_component_label(self):
        return str(self.state.x_att)

    @property
    def slice_display_unit_name(self):
        return 'spectral' if self.jdaviz_app.config == 'cubeviz' else 'temporal'

    @cached_property
    def slice_indicator(self):
        # SliceIndicatorMarks does not yet exist
        slice_indicator = SliceIndicatorMarks(self)
        self.figure.marks = self.figure.marks + slice_indicator.marks
        return slice_indicator

    @property
    def slice_values(self):
        # NOTE: these are cached at the slice-plugin level
        # Retrieve display units
        slice_display_units = self.jdaviz_app._get_display_unit(
            self.slice_display_unit_name
        )

        def _get_component(layer):
            try:
                # Retrieve layer data and units
                data_comp = layer.layer.data.get_component(self.slice_component_label)
            except (AttributeError, KeyError):
                # layer either does not have get_component (because its a subset)
                # or slice_component_label is not a component in this layer
                # either way, return an empty array and skip this layer
                return np.array([])

            # Convert axis if display units are set and are different
            data_units = getattr(data_comp, 'units', None)
            if slice_display_units and data_units and slice_display_units != data_units:
                data = np.asarray(data_comp.data, dtype=float) * u.Unit(data_units)
                return data.to_value(slice_display_units,
                                     equivalencies=u.spectral())
            else:
                return data_comp.data
        try:
            return np.asarray(np.unique(np.concatenate([_get_component(layer) for layer in self.layers])),  # noqa
                              dtype=float)
        except ValueError:
            # NOTE: this will result in caching an empty list
            return np.array([])

    def _set_slice_indicator_value(self, value):
        # this is a separate method so that viewers can override and map value if necessary
        # NOTE: on first call, this will initialize the indicator itself
        self.slice_indicator.value = value


class WithSliceSelection:
    @property
    def slice_index(self):
        # index in state.slices corresponding to the slice axis
        return 2

    @property
    def slice_component_label(self):
        slice_plg = self.jdaviz_helper.plugins.get('Slice', None)
        if slice_plg is None:  # pragma: no cover
            raise ValueError("slice plugin must be activated to access slice_component_label")
        return slice_plg._obj.slice_indicator_viewers[0].slice_component_label

    @property
    def slice_display_unit_name(self):
        return 'spectral' if self.jdaviz_app.config == 'cubeviz' else 'temporal'

    @property
    def slice_values(self):
        # NOTE: these are cached at the slice-plugin level
        # TODO: add support for multiple cubes (but then slice selection needs to be more complex)
        # if slice_index is 0, then we want the equivalent of [:, 0, 0]
        # if slice_index is 1, then we want the equivalent of [0, :, 0]
        # if slice_index is 2, then we want the equivalent of [0, 0, :]
        take_inds = [2, 1, 0]
        take_inds.remove(self.slice_index)
        converted_axis = np.array([])

        # Retrieve display units
        slice_display_units = self.jdaviz_app._get_display_unit(
            self.slice_display_unit_name
        )

        for layer in self.layers:
            coords = layer.layer.data.coords
            if hasattr(coords, 'temporal_wcs'):
                data_units = coords.temporal_wcs.unit
                data = coords.temporal_wcs.pixel_to_world_values(
                    np.arange(layer.layer.data.shape[0])
                )

            else:
                world_comp_ids = layer.layer.data.world_component_ids

                if not len(world_comp_ids):
                    # rampviz uses coordinate components:
                    world_comp_ids = layer.layer.data.coordinate_components

                if self.slice_index >= len(world_comp_ids):
                    # Case where 2D image is loaded in image viewer
                    continue

                try:
                    # Retrieve layer data and units using the slice index
                    # of the world components ids
                    data_comp = layer.layer.data.get_component(
                        world_comp_ids[self.slice_index]
                    )
                except (AttributeError, KeyError):
                    continue

                data = np.asarray(
                    data_comp.data.take(0, take_inds[0]).take(0, take_inds[1]),
                    dtype=float)
                data_units = getattr(data_comp, 'units', None)

            # Convert to display units if applicable
            if slice_display_units and data_units and slice_display_units != data_units:
                converted_axis = (data * u.Unit(data_units)).to_value(
                    slice_display_units,
                    equivalencies=u.spectral() + u.pixel_scale(1*u.pix)
                )
            else:
                converted_axis = data

        return converted_axis

    @property
    def slice(self):
        return self.state.slices[self.slice_index]

    @slice.setter
    def slice(self, slice):
        # NOTE: not intended for user-access - this should be controlled through the slice plugin
        # in order to sync with all other viewers/slice indicators
        slices = [0, 0, 0]
        slices[self.slice_index] = slice
        self.state.slices = tuple(slices)

    @property
    def slice_value(self):
        return self.slice_values[self.slice]

    @slice_value.setter
    def slice_value(self, slice_value):
        # NOTE: not intended for user-access - this should be controlled through the slice plugin
        # in order to sync with all other viewers/slice indicators
        # find the slice nearest slice_value
        slice_values = self.slice_values
        if not len(slice_values):
            return
        self.slice = np.argmin(abs(slice_values - slice_value))



from jdaviz.configs.specviz import Specviz
from jdaviz.core.helpers import ConfigHelper
from jdaviz.core.events import SnackbarMessage
from jdaviz.configs.default.plugins.line_lists.line_list_mixin import LineListMixin
from jdaviz.configs.specviz2d.plugins.spectral_extraction.spectral_extraction import SpectralExtraction  # noqa

__all__ = ['Specviz2d']


class Specviz2d(ConfigHelper, LineListMixin):
    """Specviz2D Helper class"""

    _default_configuration = "specviz2d"
    _default_spectrum_viewer_reference_name = "spectrum-viewer"
    _default_spectrum_2d_viewer_reference_name = "spectrum-2d-viewer"
    _default_table_viewer_reference_name = "table-viewer"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def specviz(self):
        """
        A Specviz helper (`~jdaviz.configs.specviz.helper.Specviz`) for the Jdaviz
        application that is wrapped by Specviz2d.
        """
        if not hasattr(self, '_specviz'):
            self._specviz = Specviz(app=self.app)
        return self._specviz

    def load_data(self, spectrum_2d=None, spectrum_1d=None, spectrum_1d_label=None,
                  spectrum_2d_label=None, show_in_viewer=True, ext=1,
                  transpose=False, cache=None, local_path=None, timeout=None):
        """
        Load and parse a pair of corresponding 1D and 2D spectra.

        Parameters
        ----------
        spectrum_2d: str
            A spectrum as translatable container objects (e.g.,
            ``Spectrum1D``) that can be read by glue-jupyter. Alternatively,
            can be a string file path.

        spectrum_1d: str or Spectrum1D
            A spectrum as translatable container objects (e.g.,
            ``Spectrum1D``) that can be read by glue-jupyter. Alternatively,
            can be a string file path.

        spectrum_1d_label : str
            String representing the label for the data item loaded via
            ``spectrum_1d``.

        spectrum_2d_label : str
            String representing the label for the data item loaded via
            ``spectrum_2d``.

        show_in_viewer : bool
            Show data in viewer(s).

        ext : int, optional
            Extension of the input ``spectrum_2d`` file to load. Defaults to 1.

        transpose : bool, optional
            Flag to transpose the 2D data array before loading. Useful for uncalibrated
            data that is dispersed vertically, to change it to horizontal dispersion.

        cache : None, bool, or str
            Cache the downloaded file if the data are retrieved by a query
            to a URL or URI.

        local_path : str, optional
            Cache remote files to this path. This is only used if data is
            requested from `astroquery.mast`.

        timeout : float, optional
            If downloading from a remote URI, set the timeout limit for
            remote requests in seconds (passed to
            `~astropy.utils.data.download_file` or
            `~astroquery.mast.Conf.timeout`).
        """
        if spectrum_2d is None and spectrum_1d is None:
            raise ValueError('Must provide spectrum_2d or spectrum_1d but none given.')

        if spectrum_2d_label is None:
            spectrum_2d_label = "Spectrum 2D"
        elif spectrum_2d_label[-2:] != "2D":
            spectrum_2d_label = "{} 2D".format(spectrum_2d_label)

        if spectrum_1d_label is None:
            spectrum_1d_label = spectrum_2d_label.replace("2D", "1D")

        load_1d = False

        if spectrum_2d is not None:
            self.app.load_data(spectrum_2d, parser_reference="mosviz-spec2d-parser",
                               data_labels=spectrum_2d_label,
                               show_in_viewer=False, add_to_table=False,
                               ext=ext, transpose=transpose,
                               cache=cache, local_path=local_path, timeout=timeout)

            # Passing show_in_viewer into app.load_data does not work anymore,
            # so we force it to show here.
            if show_in_viewer:
                self.app.add_data_to_viewer(
                    self._default_spectrum_2d_viewer_reference_name, spectrum_2d_label
                )
            # Collapse the 2D spectrum to 1D if no 1D spectrum provided
            if spectrum_1d is None:
                # create a non-interactive (so that it does not create duplicate marks with the
                # plugin-instance created later) instance of the SpectralExtraction plugin,
                # and use the defaults to generate the initial 1D extracted spectrum
                spext = SpectralExtraction(
                    app=self.app, interactive=False,
                    spectrum_viewer_reference_name=self._default_spectrum_viewer_reference_name,
                    spectrum_2d_viewer_reference_name=(
                        self._default_spectrum_2d_viewer_reference_name
                    )
                )
                # for some reason, the trailets are resetting to their default values even
                # though _trace_dataset_selected was called internally to set them to reasonable
                # new defaults.  We'll just call it again manually.
                spext._trace_dataset_selected()
                try:
                    spext.export_extract_spectrum(add_data=True)
                except Exception:
                    msg = SnackbarMessage(
                        "Automatic spectrum extraction failed. See the spectral extraction"
                        " plugin to perform a custom extraction",
                        color='error', sender=self, timeout=10000)
                else:
                    # Warn that this shouldn't be used for science
                    msg = SnackbarMessage(
                        "The extracted 1D spectrum was generated automatically."
                        " See the spectral extraction plugin for details or to"
                        " perform a custom extraction.",
                        color='warning', sender=self, timeout=10000)
                self.app.hub.broadcast(msg)

            else:
                load_1d = True

        else:
            load_1d = True

        if load_1d:
            self.app.load_data(
                spectrum_1d, data_label=spectrum_1d_label,
                parser_reference="specviz-spectrum1d-parser",
                show_in_viewer=show_in_viewer,
                cache=cache,
                local_path=local_path,
                timeout=timeout
            )

    def load_trace(self, trace, data_label, show_in_viewer=True):
        """
        Load a trace object and load into the spectrum-2d-viewer

        Parameters
        ----------
        trace : Trace
            A specreduce trace object
        data_label : str
            String representing the label
        show_in_viewer : bool
            Whether to load into the spectrum-2d-viewer.
        """
        self.app.add_data(trace, data_label=data_label)
        if show_in_viewer:
            self.app.add_data_to_viewer(
                self._default_spectrum_2d_viewer_reference_name, data_label
            )

    def get_data(self, data_label=None, spectral_subset=None,
                 cls=None, use_display_units=False):
        """
        Returns data with name equal to data_label of type cls with subsets applied from
        spectral_subset.

        Parameters
        ----------
        data_label : str, optional
            Provide a label to retrieve a specific data set from data_collection.
        spectral_subset : str, optional
            Spectral subset applied to data.
        cls : `~specutils.Spectrum1D`, `~astropy.nddata.CCDData`, optional
            The type that data will be returned as.
        use_display_units : bool, optional
            Specify whether the returned data is in native units or the current display units.

        Returns
        -------
        data : cls
            Data is returned as type cls with subsets applied.

        """
        return self._get_data(data_label=data_label, spectral_subset=spectral_subset,
                              cls=cls, use_display_units=use_display_units)
